// Example_mid.cpp 
// Sample code explaining how to use the CDirectMidi class library (Midi part)
// Last update April 27th, 2004
// Programmed by Carlos Jimnez de Parga
// Feel free to modify this code 
// Your comments and suggestions are always welcomed
// modified Nov 2006 by STB to be midi-trigger wrapper

// Main headers

// ANSI I/0 headers
#include <conio.h>
#include <iostream>
#include <math.h>
#include <assert.h>

// The class wrapper
#include ".\\DirectMidi\\CDirectMidi.h"
// Inline library inclusion

#pragma comment (lib,"dxguid.lib") // guid definitions 
#pragma comment (lib,"winmm.lib")
#pragma comment (lib,"dsound.lib")
#pragma comment (lib,"dxerr9.lib")

using namespace std;
using namespace directmidi;

// Maximum size for SysEx data in input port

const int SYSTEM_EXCLUSIVE_MEM = 48000;

// Defines PI

const double PI = 3.1415926;

static CDirectMusic CDMusic;
	//CDMReceiver  Receiver;
static COutputPort  COutPort;
static CDLSLoader   CLoader;
static CCollection  CCollectionA,CCollectionB;
static CSampleInstrument CSample1,CSample2;

#define MAX_PATCHES  32
static int patch[MAX_PATCHES];
static CInstrument program[MAX_PATCHES];
static int in_use;

int midi_make_patch(int n)
{
   int i;
   for (i=0; i < in_use; ++i) {
      if (patch[i] == n) return i;
   }
   CCollectionB.GetInstrument(program[i], n);
   program[i].SetNoteRange(0,127);
   program[i].SetPatch(i);
   patch[i] = n;
   COutPort.DownloadInstrument(program[i]);
   ++in_use;
   assert(i+1 == in_use);
   return i;
}

void midi_set_patch(int channel, int patch)
{
	COutPort.SendMidiMsg(COutputPort::EncodeMidiMsg(PATCH_CHANGE,channel,patch,0),0);
}

void midi_note_on(int channel, int note, int velocity)
{
	COutPort.SendMidiMsg(COutputPort::EncodeMidiMsg(NOTE_ON,channel,note,velocity),0); 
}

void midi_note_off(int channel, int note)
{
    COutPort.SendMidiMsg(COutputPort::EncodeMidiMsg(NOTE_OFF,channel,note,0),0); 
}


int midi_init(void)
{
	try
	{
		// Initializes DirectMusic
		
		CDMusic.Initialize();
		COutPort.Initialize(CDMusic);
		INFOPORT PortInfo;
		DWORD dwPortCount = 0;
		//Software Synthesizer selection
		do
			COutPort.GetPortInfo(++dwPortCount,&PortInfo);
		while (!(PortInfo.dwFlags & DMUS_PC_SOFTWARESYNTH));

		COutPort.SetPortParams(0,0,1,0,44100);
		COutPort.ActivatePort(&PortInfo,32);

		CLoader.Initialize();
		// Loads the deafault GM/GS collection of the software synthesizer
		CLoader.LoadDLS(NULL,CCollectionB);
	
	    #if 0
		// Gets the instrument with index 214 from the CollectionB
		CCollectionB.GetInstrument(CInstrument1,227);
		CInstrument1.SetPatch(0);
		CInstrument1.SetNoteRange(0,127);
	
		// Gets the instrument with index 0 from the CollectionA
		CCollectionB.GetInstrument(CInstrument2,0);
		CInstrument2.SetPatch(1);
		CInstrument2.SetNoteRange(0,127);
		
		// Downloads the instruments to the output ports
		
		COutPort.DownloadInstrument(CInstrument1);
		COutPort.DownloadInstrument(CInstrument2);
		#endif

        #if 0		
		int a = midi_make_patch(227);
		int b = midi_make_patch(0);
		
	
		//cout << "Notification activated" << endl;
	
		// Redirects messages from source global channel 0 to destination global channel 0
		
		// **** WARNING: Don't close the application while receiving SysEx data 
		// this could hang the application since it is dumping the buffer into a 
		// for loop
	
	
		COutPort.SendMidiMsg(COutputPort::EncodeMidiMsg(PATCH_CHANGE,1,b,0),0);
		COutPort.SendMidiMsg(COutputPort::EncodeMidiMsg(PATCH_CHANGE,2,a,0),0);
		COutPort.SendMidiMsg(COutputPort::EncodeMidiMsg(NOTE_ON,1,40,127),0); 
		getch();
		COutPort.SendMidiMsg(COutputPort::EncodeMidiMsg(NOTE_OFF,1,40,0),0); 
		COutPort.SendMidiMsg(COutputPort::EncodeMidiMsg(NOTE_ON,2,60,127),0);
		
		getch();		
		COutPort.SendMidiMsg(COutputPort::EncodeMidiMsg(NOTE_OFF,2,60,0),0);
		#endif
	} 
	catch (CDMusicException& DMExcp)
	{
		cout << "\n" << DMExcp.GetErrorDescription() << "\n" << endl;
        getch();
	}
	return 0;
}
