#include "s.h"
#include "midi.h"

enum
{
   AI_swarming,
   AI_walking,
   AI_falling,
   AI_flyin,
   AI_lane_change,
};

extern void add_ripple_source(float x, float y, float amp);

#define TUT_DELAY          (256+160+64)
#define TUT_START_OFFSET   (3*32-32)

bool show_tutorial=0?true:false;
static int next_bass=17;
int player_lane, meta_phase;
int harmony_octave = 0;
double music_time;
extern float shake;
float period = 0.2f;   // seconds per 8th
static int n = -1;
static int do_drum[16],need_bass[5];
float time_left;

int bass_at(int n);

void music_start(void)
{
   music_time = 0;
   time_left = 0;
   n = -1;
   next_bass = 16;
   harmony_octave = 0;
   while (bass_at(next_bass) == 0) ++next_bass;
   memset(do_drum, 0, sizeof(do_drum));
   memset(need_bass, 0, sizeof(need_bass));
   meta_phase = 0;
}


// target is moving at V0
// bullet moves at V1
// distance to target is D
// time to target is D / (V1+V0)
// at the next beat distance to target is: (D - V0 * beat_time)
// time to target is (D - V0*beat_time)/(V1+V0)
// delta_time is -beat_time * V0/(V1+V0)
//
// so if V0 

float LANE_LENGTH=120;

Vehicle enemy       = {  2, 200,200, .85f,.85f,  1.5,1.5, .5, 100 };

typedef struct
{
    int ch;
    int note;
    int vel;
    float time_left;
} Note;

#define MAX_ACTIVE_NOTES 1024
Note notes[MAX_ACTIVE_NOTES];
int num_notes;

void on(int ch, int note, int vel)
{
    //Object *o = new Object(vec(note/6.0-1.5, 40+prand()*LANE_LENGTH,0), T_effect);
    //o->velocity = vec(0,0,25);
    //o->lifetime = 5;
    //worldAdd(o);
    midi_note_on(ch,note,vel);
}

void off(int ch, int note)
{
    //Object *o = new Object(vec(note/6.0-1.5, 40+prand()*LANE_LENGTH,0), T_effect);
    //o->velocity = vec(0,0,-25);
    //o->lifetime = 5;
    //worldAdd(o);
    midi_note_off(ch,note);
}

void play_note(int ch, int note, int vel, float dur)
{
   if (note) {
      int i;
      // if this note is already around, retrigger
      for (i=0; i < num_notes; ++i)
         if (notes[i].ch == ch && notes[i].note == note) {
            off(ch, note);
            on(ch, note, vel);
            notes[i].time_left = dur;
            return;
         }
      on(ch, note, vel);
      int n = num_notes++;
      notes[n].ch = ch;
      notes[n].note = note;
      notes[n].time_left = dur;
   }
}

void updateNotes(float dt)
{
    for (int i=0; i < num_notes; ) {
       notes[i].time_left -= dt;
       if (notes[i].time_left < 0) {
           off(notes[i].ch, notes[i].note);
           notes[i] = notes[--num_notes];
       } else {
           ++i;
       }
    }
}

Note queued[32];
int num_queued;

void flush_queue(void)
{
    int i;
    for (i=0; i < num_queued; ++i)
       play_note(queued[i].ch, queued[i].note, queued[i].vel, 0.5);
    num_queued=0;
}

void queue_note(int ch, int note, int vel)
{
    if (time_left > 0.75 * period) {
        play_note(ch,note,vel,0.5);
    } else {
        if (num_queued == 32) return;
        int n = num_queued++;
        queued[n].ch = ch;
        queued[n].note = note;
        queued[n].vel = vel;
    }
}


static int melody [13] = { 0, 2, 3, 5, 7, 9,10,12,14,15,17,19,21};
static int melody0[13] = { 3, 5, 7, 9,11,12,14,15,17,19,21,22,24};
static int melody2[13] = { 3, 7, 7, 9,10,12,15,15,19,19,21,22,24};  //  0 - 3 -  7
static int melody1[13] = { 5, 5, 7,10,10,14,14,17,17,19,22,22,26};  // 10 - 2 -  5
static int melody3[13] = { 3, 7, 7,10,10,12,15,15,19,19,22,22,24};  //  3 - 7 - 10
static int melody4[13] = { 5, 5, 7, 9,12,12,14,17,17,19,21,24,24};  //  5 - 9 - 12


#define BD 36
#define SN 38
#define S2 40
#define T1 43
#define T2 41
#define T3 45
#define T4 47
#define T5 48
#define T6 50
#define CH 42
#define CH2 44
#define OH 46
#define CLAP 39
#define CRASH1  49
#define CRASH2  55
#define CRASH3  57
#define CHINA   52
#define RBELL   51
#define RIDE2   53
#define RIDE3   59
#define COWBELL 56
#define TAMB    54

int drum_death[] =
{
   T6,T6,T5,T5,T5,T4,T4,T3,T3,T2
};

void death_sound(Object *o)
{
    switch(o->type) {
       case T_player: play_note(10,57,127,1); break;
       case T_object: {
          static int i;
          if (o->subtype == E_crate) {
             queue_note(10, TAMB, 40);
          } else {
             queue_note(10,drum_death[i],90);
          }
          i = (i+1)%10;
          break; 
       }
    }
}

static int bassx[ 8] = { 2,2,0,2,2,2,0,2 };
static int bass0[ 8] = { 2,1,0,2,2,2,0,2 };
static int bass [32] = { 1,2,0,2,3,2,0,2, 1,2,0,2,3,2,0,2, 1,2,0,2,3,2,0,2, 4,0,4,0,3,4,1,2 };
static int cd   [ 8] = { 2,      2,       2,      2,       2,      2,       4,      4,      };
static int bass2b[16] = { 1,1,0,1,1,1,0,2, 2,2,0,2,2,2,0,3 };
static int cd2b [ 8] = { 1,      1,       2,      2,       };

static int bass2[32] = { 1,1,0,1,1,1,0,2, 2,2,0,2,2,2,0,1, 2,0,3,2,3,4,1,0, 0,1,4,3,2,1,2,0 };
static int cd2  [32] = { 1,1,1,1,1,1,1,1, 2,2,2,2,2,2,2,2, 2,2,3,3,3,3,1,1, 1,1,4,4,4,4,2,2 };
                                                         //    ^^^^^^^^^^^^^^^^^^^^^^^^^^
                                                         //   by MC This Earth That I Found
                                                         //       (DJ Ocean Bomb)
static int bass3[16] = { 2,2,2,0,1,1,1,0, 3,3,3,0,4,4,4,0 };
static int cd3  [ 4] = { 2,      1,       3,      4       };
static int bass4[16] = { 4,4,4,0,3,4,4,0, 3,4,4,0,0,4,3,2 };
static int cd4  [ 4] = { 4,      4,       4,      4       };

static int M(int x, int y)
{
   int z = x % y;
   if (z < 0) z += y;
   return z;
}

#define N(x,y)   (M(x,y) >> 2)

void get_bass_and_chord(int n, int *b, int *c)
{
   int m = n;
   //m += 256 + 32 + 64 + 32 - 32;
   m -= 32;  if (m < 0) { *b = bassx[M(m,8) ]; *c = 2; return; }
   m -= 64;  if (m < 0) { *b = bass0[M(m,8) ]; *c = 2; return; }
   m -= 32;  if (m < 0) { *b = bass [M(m,8) ]; *c = 2; return; }
   if (show_tutorial) m -= 256;
   m -= 128; if (m < 0) { *b = bass [M(m,32)]; *c = cd [N(m,32)]; return; }
   m -= 32;  if (m < 0) { *b = bass2b[M(m,16)]; *c = cd2b[N(m,16)]; return; }
   m -= 32;  if (m < 0) { *b = bass2[M(m,32)]; *c = cd2[M(m,32)]; return; }
   m -= 64;  if (m < 0) { *b = bass [M(m,32)]; *c = cd [N(m,32)]; return; }
   m -= 48;  if (m < 0) { *b = bass3[M(m,16)]; *c = cd3[N(m,16)]; return; }
   m -= 16;  if (m < 0) { *b = bass4[M(m,16)]; *c = cd4[N(m,16)]; return; }

   for(;;) {
      m -= 128; if (m < 0) { *b = bass [M(m,32)]; *c = cd [N(m,32)]; return; }
      m -= 32;  if (m < 0) { *b = bass2[M(m,16)]; *c = cd2[N(m,16)]; return; }
      m -= 32;  if (m < 0) { *b = bass2[M(m,32)]; *c = cd2[N(m,32)]; return; }
      m -= 128; if (m < 0) { *b = bass [M(m,32)]; *c = cd [N(m,32)]; return; }
      m -= 48;  if (m < 0) { *b = bass3[M(m,16)]; *c = cd3[N(m,16)]; return; }
      m -= 16;  if (m < 0) { *b = bass4[M(m,16)]; *c = cd4[N(m,16)]; return; }
   }      
}

int bass_at(int n)
{
   int b,c;
   get_bass_and_chord(n, &b, &c);
   return b;
}

static int bnote[5] = { 0,28,30,33,35, };
static int drums[16] = { BD,0,CH,CH,SN,0,CH,BD,0,BD,CH,CH,SN,0,CH,SN};
static int drump[16] = {  8,0, 3, 2, 8,0, 3, 5,0, 5, 3, 2, 8,0, 3, 5};
//static int wait_time[16]={0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0 };
// number of measures to delay each note during the intro
static int wait_time[16]={0,4,1,4, 3,4,1,4, 4,1,3,4, 0,4,2,4 };

extern float powerup_left[], powerup_start[];
static bool left_trigger, right_trigger;

void fire_left(int p)
{           
    left_trigger = right_trigger = false;
    play_note(3,54 + melody[p],70,powerup_left[1] > 0 ? period * 1.85 : period*2);
    fireShot(p2, 100, 10, player_lane);
    score -= 500000 + rand();
    float t; 
    if (powerup_left[1] >= period * 1) {
       t = 0.4;
    } else {
       t = 1;
    }
    sparks(p2->loc + vec(0,-2,0), 20, vec(0,-10,0), t, 2, E_soul2);
}

extern Object *ghost;
void fire_ghost(int p)
{           
    play_note(4,54 + melody[p],65,period*1.85);
    fireShot(ghost, 100, 11, player_min+p);
    score -= 500000 + rand();
}

void fire_right(int p, int n)
{
    int *m2[] = {
        melody0,
        melody1,
        melody2,
        melody3,
        melody4,
    };
    right_trigger = false;
    int b,c;
    get_bass_and_chord(n,&b,&c);
    play_note(5, harmony_octave+12+42+m2[c][p],60, period*1.3f);
    play_note(5, 12+42+melody[p],60, period*1.3f);
    fireShot(p2, 50, 2.5, player_lane);
//    score -= 20000 + (rand()>>4);
}

#define SWARM_COUNT 300

typedef struct
{
   Object *obj;
   float   time_left;
   int     original_target_lane;
   int     final_target_lane;
   bool    is_morphing;
   float   morph_t;
   float   morph_dt;
   float   wander_t[3];
   float   amp_x[3], phase_x[3];
   float   amp_z[3], phase_z[3];
   float   z_t2, z_t;
   float   y,dy;
} Swarmer;

Swarmer swarm[SWARM_COUNT];
int num_swarm;

char *lane_melodies[] =
{
   "1030503054321",
   "1012303450678001",
   "102010302040305",
   "112233405",
   "12302340345",
   "503402301205",
   "1103305507708005",
   "00450560670860000045056067085",
   "80560340250",
   "113322443355",
   "3450678076506",
   "87656780765454305432432032345",
   "346534654340000034653476545",
   "56765000576530007653200012345",
   "00578058706570000046704760745",
   "87656780765450008765676061234",
   "56453423241",
   "0012321053453000001232105345700000123210535787530012321054321",
   "0321321321321000043243243243205435435435430000654765876545",
   "0012314253745",
   "005353424234321",
   "8675645341234",
   "7537642653154",
   "13130242403535363637378",
   "3432565478767",
   "0012314161765",
   "86756453423120008675645342313",
   "121314151617128",
   "13131302525250373748",
   "018283848586878",
};
 

static int swarm_phase, swarm_start, swarm_n;

void set_phase(int p, int s)
{
   swarm_phase = p;
   swarm_start = s;
   swarm_n     = -1;
}

int phases[8][11] =
{
   { 0,0,0,3,3, 1,1,1,1,1, 1 },
   { 0,0,0,0,0, 0,3,3,3,2, 2 },
   { 0,0,0,0,1, 1,1,1,1,1, 1 },
   { 0,0,0,1,1, 1,1,1,1,1, 2 },
};

void choose_phase(int s)
{
   int n;
   assert(s % 32 == 0);
   if (swarm_phase == -1)
      n = 1;
   else
      n = phases[swarm_phase][rand() % 11];

   set_phase(n, s);
}

Swarmer *new_swarmer(int lane, int subtype, int delay)
{
   if (num_swarm == SWARM_COUNT) return NULL;
   Swarmer *s = &swarm[num_swarm++];
   s->obj = new Object(vec(-500,0,0), T_object);
   s->obj->v = &enemy;
   worldAdd(s->obj);
   s->time_left = time_left + period * (32+delay);
   s->final_target_lane = s->original_target_lane = lane;
   s->obj->subtype = subtype;
   switch (subtype) {
      case E_bomber  : s->obj->model = get_model("arrow"); 
                       s->obj->health = 10;
                       break;
      case E_stacked : s->obj->model = get_model("dualtop"); 
                       s->obj->health = 10;
                       break;
      case E_composed: s->obj->model = get_model("concave_pair");
                       s->obj->health = 10;
                       break;
      case E_tiny    : s->obj->model = get_model("hemi");
                       s->obj->health = 1;
                       break;
      case E_powerup : s->obj->model = get_model("pongball");
                       s->obj->health = 1;
                       s->obj->turn_rate.y = 1;
                       s->obj->turn_rate.z = 1.414;
                       break;
   }

   for (int j=0; j < 3; ++j) {
      s->amp_x[j] = pow(3, j);
      s->phase_x[j] = 0;
      s->amp_z[j] = pow(4, j)*3;
      s->phase_z[j] = 0;
      s->wander_t[j] = j;
   }
   s->dy = -50;
   s->y = LANE_LENGTH - s->time_left * s->dy + s->obj->radius() + 1;
   s->z_t = 0.5;
   s->z_t2 = 0;
   return s;
}

char debug_string[512];
void swarm_gen(int time)
{
   if (time < swarm_start) return;
   ++swarm_n;
   int half_n = swarm_n >> 1;

   int etime = time;
   if (show_tutorial) etime = etime - TUT_DELAY - TUT_START_OFFSET;

   switch(swarm_phase) {
      case -1: {
         // tutorial
         int i,j, t=0;
         //Swarmer *s = new_swarmer(player_min+5, E_powerup, -3);
         //s->obj->subsubtype = 1;

         new_swarmer(player_min + 3, E_bomber, 0);
         new_swarmer(player_min + 1, E_bomber, 16);
         new_swarmer(player_min + 6, E_bomber, 32);
         for (i=0; i < 3; ++i)
            new_swarmer(player_min + 4-i, E_bomber, 48+4*i);
         for (i=0; i < 3; ++i)
            new_swarmer(player_min + 2-i, E_bomber, 68+2*i);
         
         t = 192;   
         
         new_swarmer(player_min+7, E_tiny, t);
         
         t += 32;
         
         for (i=2; i < 6; ++i)
            new_swarmer(player_min + i, E_tiny, t);
               
         for (i=0; i < 7; ++i)
            new_swarmer(player_min + i, E_tiny, t+16);

         t += 32;
         for (i=0; i < 7; ++i)
            new_swarmer(player_min + i, E_tiny, t);
            
         for (j=0; j < 6; ++j)
            for (i=2; i <= 4; i += 2)
               new_swarmer(player_min + i, E_tiny, t+j*2);
         t += j*2;
         for (i=0; i < 3; ++i)
            for (j=0; j < 3; ++j)
               new_swarmer(player_min + 2-i, E_tiny, t+i*3+j);
         t += 64;
         for (i=1; i < 7; ++i)
            new_swarmer(player_min + i, E_bomber, t);

         t += 32;
         new_swarmer(player_min  , E_bomber, t  );
         new_swarmer(player_min+2, E_bomber, t+3);
         new_swarmer(player_min+4, E_bomber, t+6);
         new_swarmer(player_min+2, E_bomber, t+9);
         new_swarmer(player_min  , E_bomber, t+12);

         new_swarmer(player_min+5, E_powerup, t+35);

         new_swarmer(player_min+5, E_composed, t+48);
         new_swarmer(player_min+3, E_composed, t+56);
         new_swarmer(player_min+1, E_composed, t+64);
         new_swarmer(player_min+3, E_composed, t+72);

         choose_phase(time + TUT_DELAY);
         break;
      }
      
      case 0: {
         int n = ACOUNT(lane_melodies);
         float z = clamp(etime / 2000.0, 0,1);
         
         // set min and max based on clamping
         int m = n*z/2, k=m;
         int top = n * (z*3/4 + 0.25);
         if (top > n) top = n;
         m = m + ((rand() >> 3) % (top-m));
         //sprintf(debug_string, "%d %d %d %f", k,m,top, z);
         char *p = lane_melodies[m];
         assert(0 <= m && m <= top && top <= n);
         int t = -((rand() >> 2) % 3) * 2;
         while (*p) {
            int n = *p - '0';
            if (n) {
               Swarmer *s = new_swarmer(player_min + n-1, E_bomber, t);
               if (s)
                  s->obj->angles.z = prand() * 2 * 3.141592;
            }
            t += 2;
            ++p;
         }
         
         if (t >= 120) t = 128+32;
         else if (t >= 90) t = 128;
         else if (t >= 60) t = 96;
         else if (t >= 32) t = 64;
         else t = 32;
         
         choose_phase(time+t);
         break;
      }

      case 1: {
         int avg = (player_min + player_max - 1) >> 1;
         int limit = 4 + 3.2 * prand() * clamp(etime/256.,0,1);
         int width = 2.8 + 2 * prand() * clamp(etime/128.,0,1);
         int m = player_min + rand() % (player_max - player_min-width);
         int n = m+width-1;
         int rmin, rmax;
         int k;
         rmin = -1;
         rmax = +1;
         if (etime > 200 && prand() < 0.3) {
            rmin = -2;
            rmax = +2;
         }
         if (m < avg && prand() < 0.5) {
            rmin = (n < avg && prand() < 0.7) ? 0 : 1;
         } else if (n > avg && prand() < 0.5) {
            rmax = (m > avg && prand() < 0.7) ? 0 : -1;
         }
         if (etime > 450 && (rmin <= 0 || rmax >= 0) && width > 2 && prand() < 0.3)
             limit += 4 + (rand() % 5);
         int i,j;
         for (i=0; i < limit; ++i) {
            for (j=0; j < width; ++j) {
               if (prand() < 0.75) {
                  Swarmer *s = new_swarmer(m+j, E_tiny, i*2);
                  if (s) {
                     s->phase_x[0] = prand() * 6;
                     s->phase_x[1] = prand() * 6;
                     s->phase_z[0] = prand() * 6;
                  }
               }
            }
            if (i == limit-1) break;
            if (rmax == rmin)
               k = rmin;
            else
               k = (rand()>>3) % (rmax - rmin) + rmin;
            if (m + k >= player_min && n + k < player_max) {
               m += k;
               n += k;
            }
         }
         // find the nearest 'resolved' pitch
         if (m <= player_min+1)
            k = player_min;
         else {
            m = (m+n) >> 1;
            if (m <= (player_min+2+player_max-1)>>1)
               k = player_min+2;
            else
               k = player_max-1;
         }
         n = prand() < 0.7 ? 14 + (rand() % 6) : limit * 2;
         m = prand() < 0.5 ? (rand() >> 3) % 3 + 1 : 1;
         for (i=0; i < m; ++i) {
            Swarmer *s = new_swarmer(k, E_tiny, n+i);
            if (s) s->wander_t[0] = 2;
         }
         if (limit > 10) {
            new_swarmer(player_min+1 + (rand() % 3)*2, E_composed, 40 + (rand() %8));
            choose_phase(time + 64);
         } else
            choose_phase(time + 32);
         break;
      }

      case 2: {
         // give the player a powerup
         Swarmer *s = new_swarmer(player_min + 1 + (rand() % 6), E_powerup, -4);
         s->obj->subsubtype = 1;

         // now give him a few rows of spread out targets
         int row[4][4],nr[4],t,choices[16], has[16];
         float z = clamp(etime/1200.0, 0,1); // these happen rarely so scale slowly
         int r = 2.9 + (0.5 + z)*prand(),i;
         assert(r <= 4);
         
         t = 0;
         for (i=0; i < r; ++i) {
            nr[i] = 2.3 + 2*z*prand();
            t += nr[i];
         }
         
         // compute the possible lanes
         for (i=0; i < 16; ++i)
            choices[i] = player_min + (i%8);
         assert(choices[10] == player_min+2);
         assert(choices[12] == player_min+4);
         choices[12] = choices[10];
         choices[10] = choices[9];
         choices[9] = player_min+2;
         choices[10] = player_min+4;
         
         // distribute the possible lanes to the slots
         if (t < 8) t = 8;
         for (i=0; i < r; ++i) {
            memset(has, 0, sizeof(has));
            for (int j=0; j < nr[i]; ++j) {
               int n,c;
               do {
                  n = (rand()>>2) % t;
                  c = choices[n];
                  ++has[c];
               } while (has[c] != 1 && has[c] < 150);

               if (has[c] == 1) {
                  row[i][j] = c;
                  choices[n] = choices[--t];
               } else {
                  nr[i] = j;
                  break;
               }
            }
         }
         
         t = 8+(rand() % 4); // time of each row
         int c = 0; // which row config
         for (; t < 110; ) {
            // display the row
            for (int j=0; j < nr[c]; ++j) {
               new_swarmer(row[c][j], E_stacked, t);
            }
            // advance time by enough to allow for shooting each one, plus padding
            t += nr[c] * 2 + 6 - 5*z;
            if (prand() < 0.5 - .25*z) t += 2 + rand() % 4;
            
            // choose a new c that's not the same
            c = (c + (rand() % (r-1))+1) % r;
         }

         choose_phase(time+128);
         break;
      }
      
      case 3:
      {
         int pattern[3],i,j;
         int m,r,t;
         float z = clamp(etime/1000.,0,1);
         r = 4.5 + 2*z*prand();
         r = clamp(r,1,6);
         if (r == 6)
            m = 1;
         else
            m = 1 + (rand() >> 2) % (6-r);
            
         for (i=0; i < 3; ++i) {
            pattern[i] = player_min + m + (rand() >> 2) % r;
            if (i > 0) {
               if (pattern[i] == pattern[i-1] && (prand() + 0.3 > z*2))
                  --i;
               if (abs(pattern[i] - pattern[i-1]) == 1 && (prand() + 0.3 > z))
                  --i;
            }
         }
         
         t = 2 - ((rand() >> 2) % 3) * 2;
         r = 1.5 + 0.5*z + (0.7 + 1.2*z)*prand();
         r = clamp(r,1,3);
         for (i=0; i < r; ++i) {
            // put down the pattern, spaced out closer than you can play them
            for (j=0; j < 3; ++j)
               new_swarmer(pattern[j], E_composed, t + j*5);
            // now put a gap between each pattern, with enough time for you to play
            t += 3*6 + 3 + 7 * (1-z);
         }
         // now put down a final note
         m = (rand() >> 2) % 3;
         m = player_min + 1 + 2*m;
         new_swarmer(m, E_composed, t);
         t += 8;

         if (t >= 120) t = 128+32;
         else if (t >= 90) t = 128;
         else if (t >= 60) t = 96;
         else if (t >= 32) t = 64;
         else t = 32;
         
         choose_phase(time+t);
         break;
      }
   }
}

void updateSwarm(float dt)
{
    int i;
    for (i=0; i < num_swarm;) {
        Swarmer *s = &swarm[i];
        Object *o = s->obj;
        s->time_left -= dt;
        if (s->time_left < 0) {
            s->obj = NULL;
            o->loc.x = s->final_target_lane;
            o->loc.y = LANE_LENGTH;
            o->loc.z = 0;
            o->velocity.x = 0;
            o->velocity.y = -20;
            if(o->subtype == E_powerup)
               o->velocity.y = -55;
            o->velocity.z = 0; 
            o->ai_state = AI_walking;
            o->ai_timer = 0.5;
            swarm[i] = swarm[--num_swarm];
            add_ripple_source(o->loc.x, o->loc.y, 0.5);
        } else {
            int j;
            float t;
            static float x_freq[3] = { 4    ,   1,       0.055f };
            static float z_freq[3] = { 7.83f,  0.93f,    0.075f };
            static float fade_time[3] = { 0.5, 1, 2 };
            // compute location based on s->original_target_lane
            vec loc;
            s->y += s->dy * dt;
            loc.x = s->original_target_lane;  // @TODO morph to final_target_lane
            loc.y = s->y;
            t = s->time_left / 0.75f;
            if (t > 1) t= 1;
            //t = 3*t*t-2*t*t*t;
            t = 1 - (1-t)*(1-t);
            loc.z = s->z_t * s->time_left + s->z_t2 * s->time_left * s->time_left + 15*t;
            for (j=0; j < 3; ++j) {
                if (s->time_left > s->wander_t[j]) {
                    float p = (s->time_left - s->wander_t[j]) / fade_time[j];
                    if (p > 1) p = 1;
                    loc.x += sin(s->time_left * x_freq[j] + s->phase_x[j]) * s->amp_x[j] * p;
                    loc.z += sin(s->time_left * z_freq[j] + s->phase_z[j]) * s->amp_z[j] * p;
                }
            }
            o->loc = loc;
            ++i;
        }
    }
}

int player_pos_old[2]; // used for positioning the ghost
int player_fired[2];   // used to decide if the ghost should fire

void updateMusic(float dt)
{
   static float wait = -60;
   static float last_sent = 0;
   int p;
   static bool left_state, right_state;
   static bool just_went;
   
   bool left_new, right_new;

   if (dt > time_left + period) {
      // this tests if AFTER subtracting dt and adding period, we would still be negative
      dt = time_left + period/2;
   }
   
   updateNotes(dt);
   updateSwarm(dt);
   last_sent += dt;

   left_new = 0 != glfwGetMouseButton(GLFW_MOUSE_BUTTON_LEFT);
   right_new = 0 != glfwGetMouseButton(GLFW_MOUSE_BUTTON_RIGHT);
   
   if (left_new && !left_state) left_trigger = true;
   if (right_new && !right_state) right_trigger = true;
   
   left_state = left_new;
   right_state = right_new;
      
   p = player_lane - player_min;
   
   music_time += dt;
   float q = pow(0.5f, (float) (dt / 0.1f));
   shake = shake * q + 1 * (1-q);
   
   time_left -= dt;
   
   if (time_left < 0) {
       time_left += period;
       n = n+1;
       flush_queue();
       int z = n % 16;
       swarm_gen(n);
       if (!(show_tutorial && music_time < TUT_DELAY)) {
         if (powerup_left[0] < wait && prand() < 0.125 && last_sent > 20) {
            new_swarmer(player_min + rand() % 8, E_powerup, -4);
            last_sent = 0;
            wait = -50 - 30 * prand();
         }
       }
      
       if (n % 2 == 0) {
          player_pos_old[1] = player_pos_old[0];
          player_pos_old[0] = player_lane;
          player_fired[1] = player_fired[0];
          player_fired[0] = 0;
       }
       if (drump[z]) {
           if (do_drum[z]) {
               play_note(10, drums[n%16], 90, 0.25f);
               shake = drump[n%16];
               ++need_bass[do_drum[z]];
           }
           do_drum[z] = 0;
           if (wait_time[z]*16*period < music_time) {
               int b = bass_at(next_bass);
               assert(b > 0);
               assert(next_bass >= n+16);
               do_drum[z] = b;
               Object *o = new Object(vec(b-1, LANE_LENGTH+85*period*16, 100), T_object);
               if (drums[z] == BD)
                  o->model = get_model("box");
               else if (drums[z] == SN)
                  o->model = get_model("snare");
               else
                  o->model = get_model("box2");
               o->ai_data2 = drums[z];
               o->subtype = E_crate;
               o->velocity.y = -85;
               o->velocity.z = -100.0 / (16*period);
               o->ai_state = AI_flyin;
               o->health = 25;
               o->v = &enemy;
               worldAdd(o);
           }
           do
              ++next_bass;
           while (bass_at(next_bass) == 0);
       }
       if (bnote[bass_at(n)]) {
           int z = bass_at(n);
           if (need_bass[z]) {
              --need_bass[z];
              play_note(2, bnote[z], 70, period*1.25f);
              play_note(1, bnote[z], 70, period*1.25f);
              fireShot(mbase, 100, 50, z-1);
           } else {
              #if 0
              Object *o = new Object(mbase->loc, T_effect);
              o->v = &enemy;
              o->velocity.y = 50;
              o->loc.y += 0;
              o->lifetime = 100;
              worldAdd(o);
              #endif
           }
       }
       if ((n % 2) == 0 && (left_new || left_trigger) ) {
          fire_left(p);
          player_fired[0] = 1;
          just_went = true;
       } else if (right_trigger || (!left_new && right_new)) {
          fire_right(p,n);
          just_went = true;
       } else {
          just_went = false;
       }
       if (n % 2 == 1) {
          if (powerup_left[1] > 0 && player_fired[1]) {
             fire_ghost(player_pos_old[1] - player_min);
          }
       }
   } else if (time_left > 0.5 * period ) {
       // check for a 'late' hit  
       if ((n % 2 == 0) && left_trigger) {
          if (just_went)
             left_trigger = right_trigger = false;
          else {
             fire_left(p);
             player_fired[0] = 1;
             just_went = true;
          }
       } else if (right_trigger) {
          if (just_went)
             right_trigger = false;
          else {
             fire_right(p,n);
             just_went = true;
          }
       }
   }
   
   float z = (period - time_left) / period;
   int b0,b1;
   b0 = bass_at(n+1);
   b1 = bass_at(n);
   if (b1 == 0) {
      int p = n-1;
      while (bass_at(p) == 0)
         --p;
      b1 = bass_at(p);
   }
   if (b0 == 0) b0 = b1;
   z = 3*z*z-2*z*z*z;
   float s = b1*(1-z) + z*b0;
   mbase->loc.x = s-1;
   mbase->loc.y = -3;
   mbase->loc.z = 0;

   // n=0:
   //  0 - 0.5    ->  player_pos_old[1]
   //  0.5 - 1    ->  player_pos_old[1]
   // n=1
   //  0 - 1      ->  from player_pos[1] to player_pos_old[0]
   if (n % 2 == 0)
      ghost->loc.x = player_pos_old[1];
   else {
      ghost->loc.x = player_pos_old[1] + z * (player_pos_old[0] - player_pos_old[1]);
   }

   
}

void dropBomb(Object *o)
{
   Object *e = new Object(o->loc, T_effect);
   e->velocity.x = frand() * 3;
   e->velocity.y = 30 + frand() * 20;
   e->velocity.z = -50;
   e->velocity *= 3;
   e->lifetime = 5;
   e->subtype = E_bomb;
   worldAdd(e);
   queue_note(10, COWBELL, 20 + (1-o->loc.y/LANE_LENGTH)*60);
}

void do_powerup(int type)
{
   static float powerup_time[] = { 64, 122 };

   powerup_left[type] = powerup_time[type] * period;
   powerup_start[type] = 0;
}

void aiTick(Object *o, float dt)
{
    if (o->subtype == E_bomber)
       o->angles.z += dt * 9;
    o->ai_timer -= dt;
    switch (o->ai_state) {
       case AI_swarming: break;
       
       case AI_walking:
           if (o->loc.y < -3) {
              if (o->subtype == E_powerup) {
                 if (o->loc.x == player_lane) {
                    do_powerup(o->subsubtype);
                    explosion(o->loc, 25, 50, 0.5, 50, 3);
                    o->kill();
                    queue_note(10,CRASH1,90);
                 } else {
                    explosion(o->loc,  5, 50, 0.5, -50, 0);
                    o->kill();
                    queue_note(10,CHINA,60);
                 }
                 break;
              }
              o->ai_state = AI_falling;
              o->velocity.y = 10 + frand() * 15;
              o->velocity.x = frand() * 10;
              o->velocity.z = -20;
              queue_note(10,58,80);
              break;
           }
           if (o->subtype == E_bomber && o->ai_timer < 0 && o->loc.y < LANE_LENGTH * 0.75f) {
              dropBomb(o);
              o->ai_timer = (o->loc.y+10) / 80;
           }
           break;
       case AI_falling:
           o->velocity.z -= 40 * dt;
           o->velocity.y += 30 * dt;
           if (o->loc.z < -900) {
               explosion(o->loc, 150, 80, 2.5, 80, E_soul);
               o->kill();
               score -= 20000000;
           }
           break;
       case AI_flyin:
           if (o->loc.z <= 0) {
               o->velocity.z = 0;
               o->loc.z = 0;
               if (o->ai_data2 == BD || o->ai_data2 == SN)
                  add_ripple_source(o->loc.x, o->loc.y, 2);
               o->ai_state = AI_walking;
               o->ai_timer = o->loc.y < LANE_LENGTH/2 ? 0 : (o->loc.y+10) / 80;
           }
           break;
       case AI_lane_change:
           if (o->velocity.x > 0 ? o->loc.x >= o->ai_data : o->loc.x <= o->ai_data) {
               o->loc.x = o->ai_data;
               o->angles = vec();
               o->velocity.x = 0;
               o->ai_state = AI_walking;
               o->turn_rate = vec();
               break;
           }
           break;
    }
}

void aiDie(Object *o)
{
    if (o->type == T_object) {
        switch(o->subtype) {
           case E_composed: {
               Object *p,*q;
               p = new Object(o->loc - vec(0.15,0,0), T_object);
               q = new Object(o->loc + vec(0.15,0,0), T_object);
               p->subtype = q->subtype = E_split;
               p->v = q->v = &enemy;
               p->model = q->model = get_model("concave");
               p->health = q->health = 10;
               p->ai_state = q->ai_state = AI_lane_change;
               p->ai_data = o->loc.x - 1;
               q->ai_data = o->loc.x + 1;
               p->velocity = q->velocity = o->velocity;
               p->velocity.x = -4;
               q->velocity.x =  4;
               p->angles.y =  3.141592/2;
               q->angles.y = -3.141592/2;
               p->turn_rate.y = -p->angles.y * 4.2;
               q->turn_rate.y = -q->angles.y * 4.2;
               worldAdd(p);
               worldAdd(q);
               break;
           }
           case E_stacked: {
               Object *p;
               p = new Object(o->loc + vec(0,0,3.5), T_object);
               p->subtype = E_bomber;
               p->v = &enemy;
               p->model = get_model("arrow");
               p->health = 10;
               p->ai_state = AI_flyin;
               p->velocity = o->velocity;
               p->velocity.z = -20;
               p->angles = o->angles;
               p->turn_rate = o->turn_rate;
               worldAdd(p);
               break;
           }
        }
    }
}

void aiInit(void)
{
   if (show_tutorial)
      set_phase(-1, TUT_START_OFFSET);
   else
      set_phase(0,32);
}
