#include <assert.h>
#include "s.h"

struct Surface
{
    char *name;
    char **lines;
    int num_lines;
    float scale_x, scale_y;
    int num_slices;
    Mesh *m;
};

#define SD(x,sx,sy,n)  { #x, x, ACOUNT(x), sx, sy, n, NULL }
#define SD2(x,s,n) SD(x,s,s,n)
#define SDX(name,x,sx,sy,n) { #name,x,ACOUNT(x),sx,sy,n,NULL }

static char *arrow[] = {
"  a                 r        ",
"   b                         ",
"      d             p        ",
"                  o          ",
"           h  m              ",
"                             ",
"       j                     ",
"                             ",
};

static char *box[] = {
"a                 z",
"b                 y",
"                   ",
"e                 w",
};

static char *tube[] = {
"a                               z",
"b  d                         w  y",
" c    fj                 qu    x ",
"                                 ",
"        i               r        ",
"      g h               s t      ",
};

static char *cannon[] = {
"a                                 z",
"                                   ",
"                                   ",
"                                   ",
"                                   ",
"                                   ",
"                                   ",
"b                                 v",
};

static char *cannon2[] = {
"a                                 z",
"                                   ",
"                                   ",
"                                 y ",
"b                       s           ",
"                          t     x  ",
"                                   ",
"                            u w    ",
};


static char *base[] = {
"a                                     n",
"    b                            m  ",
"       c                      l     ",
"                                    ",
"          d                k        ",
"                                    ",
"            e            j          ",
"                                    ",
"               f      i             ",
"                 g  h               ",
"                                    ",
};

static char *hemi[] = {
"a         j        ",
"b                  ",
"                   ",
" d                 ",
"                   ",
"    f              ",
"        h i        ",
"                   ",
};

static char *concave[] = {
"a                        s",
"                         r",
"                     q    ",
"                          ",
"        i    k     p      ",
"    g            m        ",
"                  n       ",
" e                        ",
"                          ",
"c                         ",
};


static char *ghost[] = {
"         a                   x      ",
"                                    ",
"                                    ",
"                                    ",
"       d                       u    ",
"                                    ",
"   gf                             sr",
"                                    ",
"     h                           q  ",
"                                    ",
"          j                 o       ",
"                 l   m              ",
"                                    ",
};



static char *blank[] = {
"                                   ",
"                                   ",
};

Surface surfaces[] =
{
    SD2(arrow,.4,8),
    SD2(box,.9,4),
    SDX(box2,box,.5,.5,4),
    SDX(box3,box,.3,.5,8),
    SD(tube,.45,.3,16),
    SD(cannon2,.20,.15,16),
    SD(cannon,.20,.15,16),
    SD(base,.35,.25,18),
    SD(hemi, .2,.2,12),
    SD(concave, .15,.25,5),
    SD(ghost,.10,.15,12),
    SDX(cannon3,cannon,0.30,0.10,10),
};

Mesh *get_mesh(char *name)
{
    int i;
    for (i=0; i < ACOUNT(surfaces); ++i)
        if (!stricmp(surfaces[i].name, name))
            return surfaces[i].m;
    assert(0);            
    return NULL;        
}

struct NamedModel
{
   char *name;
   Model *m;
};
vector<NamedModel> nms;

void add_model(char *name, Model *m)
{
   NamedModel z = { name, m };
   nms.push_back(z);
}

Model *get_model(char *name)
{
   for (int i=0; i < nms.size(); ++i)
      if (!stricmp(nms[i].name, name))
          return nms[i].m;
   assert(0);
   return NULL;
}

struct NamedTexture
{
   char *name;
   GLuint tex_handle;
};

vector<NamedTexture> nt;

void tex(char *name)
{
   char filename[256];
   NamedTexture n;
   n.name = strdup(name);
   sprintf(filename, "%s.png", name);
   n.tex_handle = textureLoad(filename);
   if (n.tex_handle == 0) {
      sprintf(filename, "%s.jpg", name);
      n.tex_handle = textureLoad(filename);
      if (n.tex_handle == 0) {
         assert(0);
         return;
      }
   }
   nt.push_back(n);
}

GLuint get_texture(char *name)
{
   for (int i=0; i < nt.size(); ++i)
      if (!stricmp(nt[i].name, name))
         return nt[i].tex_handle;
   assert(0);
   return 0;
}

static void build_icos(Mesh *m, float r, float protrude);

void init_artwork(void)
{
    int i;
    tex("lines");
    tex("top");
    tex("claw");
    tex("base");
    tex("crate");
    tex("crate2");
    tex("crate3");
    tex("top2");
    for (i=0; i < ACOUNT(surfaces); ++i) {
        surfaces[i].m = revolute_surface(surfaces[i].lines, surfaces[i].num_lines, surfaces[i].scale_x, surfaces[i].scale_y, surfaces[i].num_slices);  
        surfaces[i].m->optimize();
    }    
    
    Transform t;
    Model *m;
    
    m = new Model;
    t.ori = rotationMatrix(-3.141592/2, vec(0,0,1));
    t.ori = t.ori * rotationMatrix(-3.141592/4, vec(1,0,0));
    m->addMesh("box", "crate", &t);
    add_model("box", m);
    
    m = new Model;
    t.ori = rotationMatrix(-3.141592/2, vec(0,0,1));
    t.ori = t.ori * rotationMatrix(-3.141592/4, vec(1,0,0));
    m->addMesh("box", "crate2", &t);
    add_model("snare", m);

    m = new Model;
    t.ori = rotationMatrix(-3.141592/2, vec(0,0,1));
    t.ori = t.ori * rotationMatrix(-3.141592/4, vec(1,0,0));
    m->addMesh("box2", "crate3", &t);
    add_model("box2", m);
    
    m = new Model;
    t.loc = vec();
    t.ori = rotationMatrix(3.141592/2, vec(0,1,0));
    m->addMesh("arrow", "top", &t);
    add_model("arrow", m);

    m = new Model;
    t.loc = vec(0,0,1);
    t.ori = rotationMatrix(-3.141592/2, vec(0,1,0));
    m->addMesh("box3", "crate2", &t);
    t.loc = vec(0,0,3.5);
    t.ori = rotationMatrix(3.141592/2, vec(0,1,0));
    m->addMesh("arrow", "top", &t);
    add_model("dualtop", m);
    
    m = new Model;
    m->addMesh("tube", "claw");
    add_model("tube", m);

    m = new Model;
    t.loc = vec();
    t.ori = rotationMatrix(-3.141592/2, vec(0,0,1));
    m->addMesh("cannon2", "base",&t);
    add_model("cannon", m);    

    m = new Model;
    m->addMesh("base", "base");
    t.loc = vec(0,2,0);
    t.ori = rotationMatrix(-3.141592/2, vec(0,0,1));
    m->addMesh("cannon", "base",&t);
    add_model("base", m);
    
    m = new Model;
    t.loc = vec();
    t.ori = rotationMatrix(3.141592/2, vec(0,1,0));
    m->addMesh("hemi", "top2", &t);
    add_model("hemi", m);

    m = new Model;
    t.loc = vec();
    t.ori = rotationMatrix(-3.141592/2, vec(0,1,0));
    m->addMesh("concave", "top2", &t);
    add_model("concave", m);

    m = new Model;
    t.loc = vec(-0.9,0,0);
    t.ori.identity();
    m->addMesh("concave", "top2", &t);
    t.loc = -1 * t.loc;
    t.ori = rotationMatrix(-3.141592, vec(0,1,0));
    m->addMesh("concave", "top2", &t);
    add_model("concave_pair", m);

    Mesh *n = new Mesh;
    build_icos(n, 1.75, 0.8);
    m = new Model;
    m->addMesh(n, 0);
    add_model("pongball", m);

    m = new Model;
    t.loc = vec(0,2,0);
    t.ori = rotationMatrix(-3.141592/2, vec(0,0,1));
    m->addMesh("cannon3", "base", &t);
    add_model("ghost", m);
    m->addMesh("ghost", "base");
    m->addMesh("ghost", "base"); // draw it twice for awesomeness
}

#define GM  1.618034
static float verts_old[3][4][3] =
{ 
  {
   {  GM, 1,0 },
   { -GM, 1,0 },
   {  GM,-1,0 },
   { -GM,-1,0 },
  },

  {
   { 1,0, GM },
   { 1,0,-GM},
   {-1,0, GM},
   {-1,0,-GM},
  },

  {
   {0, GM, 1},
   {0,-GM, 1},
   {0, GM,-1},
   {0,-GM,-1}
  },
};

static vec verts[3][4];

static void draw_t(Mesh *m, vec &p0, vec &p1, vec &p2)
{
   vec n,d0,d1;
   d0.x = p1.x - p0.x;
   d0.y = p1.y - p0.y;
   d0.z = p1.z - p0.z;
   d1.x = p2.x - p1.x;
   d1.y = p2.y - p1.y;
   d1.z = p2.z - p1.z;
   n.x = d0.y*d1.z - d1.y*d0.z;
   n.y = d0.z*d1.x - d1.z*d0.x;
   n.z = d0.x*d1.y - d1.x*d0.y;
   n.normalize();
   m->addTriangle(m->addVertex(p0, n, 0,0),
                  m->addVertex(p1, n, 0,0),
                  m->addVertex(p2, n, 0,0));
}

float global_r, Pong_ball_protrude;

static void build_new(Mesh *m, vec &a, vec &p0, vec &p1)
{
   a = p0+p1;
   a.normalize();
   a *= global_r;
}

static void t(Mesh *m, vec &p0, vec &p1, vec &p2, int r)
{
   if (!r) {
      draw_t(m,p0,p1,p2);
   } else if (r == 1) {
      float prot = Pong_ball_protrude;
      vec p;
      p.x = p0.x + p1.x + p2.x;
      p.y = p0.y + p1.y + p2.y;
      p.z = p0.z + p1.z + p2.z;
      p.normalize();
      p.x *= global_r * prot;
      p.y *= global_r * prot;
      p.z *= global_r * prot;
      t(m,p0,p1,p,r-1);
      t(m,p1,p2,p,r-1);
      t(m,p2,p0,p,r-1);
   } else {
      vec p01,p12,p20;
      build_new(m,p01,p0,p1);
      build_new(m,p12,p1,p2);
      build_new(m,p20,p2,p0);
      t(m,p1,p12,p01,r-1);
      t(m,p2,p20,p12,r-1);
      t(m,p0,p01,p20,r-1);
      t(m,p01,p12,p20,r-1);
   }
}

#define R 2

#define p11   verts[0][0]
#define p12   verts[0][1]
#define p13   verts[0][2]
#define p14   verts[0][3]

#define p21   verts[1][0]
#define p22   verts[1][1]
#define p23   verts[1][2]
#define p24   verts[1][3]

#define p31   verts[2][0]
#define p32   verts[2][1]
#define p33   verts[2][2]
#define p34   verts[2][3]

static void build_icos(Mesh *m, float r, float protrude)
{
   memcpy(verts, verts_old, sizeof(verts));
   Pong_ball_protrude = protrude; // 0.8

   global_r = r;

   int i;
   for (i=0; i < 12; ++i) {
      float x = verts[0][i].x;
      float y = verts[0][i].y;
      float z = verts[0][i].z;
      float d = 1.0 / sqrt(x*x + y*y + z*z);
      verts[0][i].x *= d * r;
      verts[0][i].y *= d * r;
      verts[0][i].z *= d * r;
   }
   t(m,p11,p31,p21,R);
   t(m,p11,p22,p33,R);
   t(m,p13,p21,p32,R);
   t(m,p13,p34,p22,R);
   t(m,p12,p23,p31,R);
   t(m,p12,p33,p24,R);
   t(m,p14,p32,p23,R);
   t(m,p14,p24,p34,R);
   
   t(m,p11,p33,p31,R);
   t(m,p12,p31,p33,R);
   t(m,p13,p32,p34,R);
   t(m,p14,p34,p32,R);

   t(m,p21,p13,p11,R);
   t(m,p22,p11,p13,R);
   t(m,p23,p12,p14,R);
   t(m,p24,p14,p12,R);

   t(m,p31,p23,p21,R);
   t(m,p32,p21,p23,R);
   t(m,p33,p22,p24,R);
   t(m,p34,p24,p22,R);
   m->optimize();
}
