#include <math.h>
#include <time.h>
#include "s.h"
#include "midi.h"

double computeFrameTime()
{
   static double prev_time = -1;
   double cur_time = glfwGetTime();
   double delta_time;

   if (prev_time < 0)
      delta_time = 1.0 / 60; // 1/60th of a second
   else
      delta_time = cur_time - prev_time;

   prev_time = cur_time;
   return delta_time;
}

void processFrame()
{
   float dt = computeFrameTime();
   worldTick(dt);
   
   glClearColor(0,0,0,1);    // blue
   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

   worldDraw();
}

void setup_patches(void)
{
    midi_set_patch( 1, midi_make_patch(75));
    midi_set_patch( 2, midi_make_patch(76));   // SynthBass101  -- 75 is also good
    midi_set_patch( 3, midi_make_patch(101));  // trumpet--no actually trombone
    midi_set_patch( 4, midi_make_patch(101));  // fake echo needs its own channel
    midi_set_patch( 5, midi_make_patch(13));   // EP
    midi_set_patch(10, midi_make_patch(226));  // Standard drums
    midi_set_patch(11, midi_make_patch(184));  // reverse cymbal
}

bool fullscreen = true;
int main(int argc, char **argv)
{
   int i;
   for (i = 1; i < argc; ++i) {
      if (!strcmp(argv[i], "-windowed")) fullscreen = false;
      else if (!strcmp(argv[i], "-fullscreen")) fullscreen = true;
   }
   glfwInit();
   if (fullscreen)
   {
      GLFWvidmode all_modes[256];
      // make sure we can get our fullscreen mode
      int n = glfwGetVideoModes(all_modes, sizeof(all_modes)/sizeof(all_modes[0]));
      for (i=0; i < n; ++i) {
         if (all_modes[i].Width == 1024 && all_modes[i].Height == 768)
            break;
      }
      if (i == n)
         fullscreen = false;
   }
  retry:
   glfwOpenWindow(1024,768, 8,8,8,8,32,0,fullscreen ? GLFW_FULLSCREEN : GLFW_WINDOW);
   if (fullscreen) {
      int sx,sy;
      glfwGetWindowSize(&sx,&sy);
      if (sx != 1024 || sy != 768) {
         fullscreen = false;
         goto retry;
      }
   }
   glfwDisable(GLFW_MOUSE_CURSOR);
   srand(time(NULL));
   midi_init();
   setup_patches();
   
   //glfwEnable(GLFW_STICKY_MOUSE_BUTTONS);

   glfwSwapInterval(1);
   worldInit();
   while(glfwGetWindowParam(GLFW_OPENED) && !glfwGetKey(GLFW_KEY_ESC)) {
      processFrame();
      glfwSwapBuffers();
   }
   return 0;
}
