#include "s.h"

// how big should the map be?  at 200 f/s max non-afterburner speed,
// if we want it to take 60 seconds to cross the map, we need it to
// be 200*60 = 12000 feet. At 128x128, that's 100 ft per map square
// (which is pretty huge). let's go for 50ft/square instead


typedef struct
{
   float x,y;
   float time, amplitude;
} RippleSource;

#define MAX_SOURCES  7
RippleSource sources[MAX_SOURCES];
int num_ripple;

void add_ripple_source(float x, float y, float amp)
{
   int i;
   if (num_ripple == MAX_SOURCES) {
       int j;
       i = 0;
       for (j=1; j < MAX_SOURCES; ++j)
          if (sources[j].time > sources[i].time)
             i = j;
   } else
       i = num_ripple++;
   sources[i].x = x * LANE_SPACING;
   sources[i].y = y;
   sources[i].amplitude = amp;
   sources[i].time = 0;
}

#define RIPPLE_TIME 5
void update_ripples(float dt)
{
   for (int i=0; i < num_ripple; ) {
      sources[i].time += dt;
      if (sources[i].time > RIPPLE_TIME)
         sources[i] = sources[--num_ripple];
      else
          ++i;
   }
}

const float sea_level = -400;
GLuint line_tex;

#define MAPMASK   (MAPSIZE-1)
float height[MAPSIZE][MAPSIZE];

// size of the entire map in world units
extern double music_time;

float mapHeight(float x, float y)
{
   return 0;
}

float compute_map_height(float x, float y)
{
   int i;
   float z = 0;
   //float c = cos(music_time*8 + y*2) + cos(music_time*8 + y*2*shake) * (shake-1)/2;
   for (i=0; i < num_ripple; ++i) {
      float dx = (sources[i].x - x);
      float dy = sources[i].y - y;
      float dist = sqrt(dx*dx + dy*dy);
      float d = sources[i].time*24 - dist/4;
      float h = sin(-d);
      d = d / 15;
      if (d > 0) {
         // damp with distance
         float e = dist / 140;
         if (e > 1) e = 1;
         e = 1-e;
         // wave gets smaller quickly
         if (d > 1) d = 1;
         d = 1-d;
         d = d*d*d;
         // damp sideways
         dx = fabs(dx);
         dx = 1 - dx/30;
         if (dx < 0) dx = 0;
         dx = 3*dx*dx - 2*dx*dx*dx;
         h *= 1.5;
         z += h * d * sources[i].amplitude * dx * e;
      }
   }
   return z;
}

float shake=1;
void vertex(float x, float y)
{
   float a = sin(music_time*4 + x/4) * shake / 4;
   float b = sin(music_time*20 + y/2 + x/5) + sin(music_time*20 + y/2*shake) * (shake-1)/2;
   glVertex3f(x+b/8,y+a/4,compute_map_height(x,y));
}

void drawLane(float x0, float y0, float x1, float y1, float colors[32][4])
{
   int i,n;
   float p=0, dp = 1.0f / 31;
   p = 0;
   i = 0;
   for (n=31; n >= 0; --n)
       if (colors[n][0] || colors[n][1] || colors[n][2] || colors[n][3]) 
           break;
   if (n < 0) return;
   for (i=0; i < n; ++i, p += dp)
       if (colors[i][0] || colors[n][1] || colors[n][2] || colors[n][3])
           break;
   if (i > 0) --i;
   if (n < 31) ++n;
   glBegin(GL_QUAD_STRIP);
      for(; i <= n; ++i, p += dp) {
         glColor4fv(colors[i]);
         vertex(x0,y0+(y1-y0)*p);
         vertex(x1,y0+(y1-y0)*p);
      }
   glEnd();
}

void drawRectY(float x0, float y0, float x1, float y1, float s, float t)
{
   float p;
   glBegin(GL_QUAD_STRIP);
      for(p=0; p <= 1; p += 1.0f/256) {
         glTexCoord2f(0,p*t); vertex(x0,y0+(y1-y0)*p);
         glTexCoord2f(s,p*t); vertex(x1,y0+(y1-y0)*p);
      }
   glEnd();
}

void drawRectX(float x0, float y0, float x1, float y1, float s, float t)
{
   float p;
   glBegin(GL_QUAD_STRIP);
      for(p=0; p <= 1; p += 1.0f/16) {
         glTexCoord2f(0,p*t); vertex(x0+(x1-x0)*p, y1);
         glTexCoord2f(s,p*t); vertex(x0+(x1-x0)*p, y0);
      }
   glEnd();
}

extern player_lane;
float player_lane_frac;

float lane_colors[32][32][4];

void mapDraw(void)
{
   int i;
   glEnable(GL_BLEND);
   glDisable(GL_DEPTH_TEST);
   glDisable(GL_TEXTURE_2D);
   glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
   for (i=0; i < NUM_LANES; ++i) {
      drawLane(i*LANE_SPACING+BAND_WIDTH/2,0,(i+1)*LANE_SPACING+BAND_WIDTH/2,LANE_LENGTH,lane_colors[i]);
   }
   glBindTexture(GL_TEXTURE_2D, line_tex);
   glEnable(GL_TEXTURE_2D);
   for (i=0; i <= NUM_LANES; ++i) {
      if (i == player_lane || i == player_lane+1) {
          float p;
          if (i == player_lane) p = 1-player_lane_frac;
          else p = player_lane_frac;
          p = 1 - (1-p)*(1-p);
          p = 1 - (1-p)*0.75;
          glColor3f(0.35f+p*0.65f, 0.35f+p*0.65f, 0.55f - 0.3f*p);
      } else {
          glColor3f(0.35f,0.35f,0.65f);
      }
      drawRectY(i*LANE_SPACING,0, i*LANE_SPACING+BAND_WIDTH,LANE_LENGTH,1,LANE_LENGTH/BAND_WIDTH);
   }
   glColor3f(0.35f,0.35f,0.65f);
   drawRectX(0,-BAND_WIDTH/2,NUM_LANES*LANE_SPACING+BAND_WIDTH,BAND_WIDTH/2,1,NUM_LANES*LANE_SPACING/BAND_WIDTH);
   drawRectX(0,LANE_LENGTH-BAND_WIDTH*2,NUM_LANES*LANE_SPACING+BAND_WIDTH,LANE_LENGTH,1,NUM_LANES*LANE_SPACING/BAND_WIDTH);
   glDisable(GL_BLEND);
   glEnable(GL_DEPTH_TEST);
}

#define NUM_BASE_LOCATIONS 4

struct
{
   int x,y;
   float h;
} baseloc[NUM_BASE_LOCATIONS];

// base size in map units
#define BASE_SIZE 4

static float minHeight(int x, int y)
{
   float h=height[y][x];
   for (int dy=-BASE_SIZE; dy <= BASE_SIZE; ++dy)
   for (int dx=-BASE_SIZE; dx <= BASE_SIZE; ++dx)
      if (height[y+dy][x+dx] < h)
         h = height[y+dy][x+dx];
   return h;
}

void mapInit(void)
{
   int i;
   uint8 data[16][4];
   const float trans     =  100;
   
   for (i=0; i < 16; ++i) {
      float p = 1 - abs(i - 7.5f)/7.5f;
      float q = 3*p*p - 2*p*p*p;
      float r = q*q;
      data[i][0] = 255*q;
      data[i][1] = 255*q;
      data[i][2] = 255*q;
      data[i][3] = 255*q;
   }
   
   line_tex = texCreate(16,1,data[0]);
   textureClamp(line_tex);
   
   // now make another fractal to use for texturing
   //map_tex = createMapTexture(512);
   //map_list = glGenLists(1);
   //glNewList(map_list, GL_COMPILE);
   //heightFieldDraw();
   //glEndList();
}

