#include "s.h"

// let's replace this later, rather than overloading our source delta right now
int Mesh::addVertex(const vec &loc, const vec &norm, float s, float t)
{
   for (int i=0; i < vertices.size(); ++i) {
      const Vertex &v = vertices[i];
      if (loc == v.loc && norm == v.norm && s == v.s && s == v.t)
         return i;
   }
   Vertex v; v.loc = loc; v.norm = norm; v.s = s; v.t = t;
   vertices.push_back(v);
   return vertices.size()-1;
}

void Model::draw()
{
   glEnable(GL_TEXTURE_2D);
   for (int i=0; i < meshes.size(); ++i) {
      glBindTexture(GL_TEXTURE_2D, textures[i]);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      float temp[16];
      glmat(temp, xforms[i].ori, xforms[i].loc);
      glMultMatrixf(temp);
      meshes[i]->draw();
      glMatrixMode(GL_MODELVIEW);
      glPopMatrix();
   }
   glDisable(GL_TEXTURE_2D);
}

void Mesh::draw()
{
   if (display_list) {
      glCallList(display_list);
      return;
   }
   int stride = sizeof(vertices[0]);

   glVertexPointer(3, GL_FLOAT, stride, &vertices[0].loc.x);
   glEnableClientState(GL_VERTEX_ARRAY);
   glNormalPointer      (GL_FLOAT, stride, &vertices[0].norm.x);
   glEnableClientState  (GL_NORMAL_ARRAY);
   glTexCoordPointer(2, GL_FLOAT, stride, &vertices[0].s);
   glEnableClientState(GL_TEXTURE_COORD_ARRAY);
   glDrawElements(GL_TRIANGLES, indices.size(), GL_UNSIGNED_INT, &indices[0]);
   glDisableClientState(GL_VERTEX_ARRAY);
   glDisableClientState(GL_TEXTURE_COORD_ARRAY);
   glDisableClientState(GL_NORMAL_ARRAY);
}

void Mesh::optimize()
{
   int stride = sizeof(vertices[0]);
   if (display_list) {
      glDeleteLists(display_list, 1);
      display_list = 0;
   }
   GLuint temp_display_list = glGenLists(1);
   glNewList(temp_display_list, GL_COMPILE);
   draw();
   glEndList();
   display_list = temp_display_list;
}


int get_2d_shape(vec *dest, char **src, int n_src)
{
    int n=0;
    int w = strlen(src[0]) / 2;
    for (int i='a'; i <= 'z'; ++i) {
        for (int j=0; j < n_src; ++j) {
            char *s = strchr(src[j], i);
            if (s) {
                dest[n].x = (s - src[j] - w)*0.75f;
                dest[n].y = j;
                dest[n].z = i - 'a'; // s texture coordinate
                ++n;
                break;
            }
        }
    }
    // renormalize to all texture coordinates
    for (i=0; i < n; ++i)
        dest[i].z /= dest[n-1].z;
    return n;
}

Mesh *revolute_surface(char **src, int n_src, float scale_x, float scale_y, int slices)
{
   Mesh *m;
   int i,j=0,n,k;
   vec raw[27],surf[27];
   vec pts[55],norm[55];
   int initial[55],cur[55],prev[55];
   n = get_2d_shape(raw, src, n_src);
   if ( n < 3) return NULL;

   for (i=0; i < n; ++i)
		 raw[i].x *= scale_x, raw[i].y *= scale_y;

   for (i=0; i < n-1; ++i) {
       vec delta = raw[i+1] - raw[i];
       delta.normalize();
       surf[i].x = -delta.y;
       surf[i].y =  delta.x;
       surf[i].z = 0;
   }
   
   k = 0;
   i = 0;
   pts [k] = raw [i];
   norm[k] = vec(-1,0,0);
   ++i,++k;
   for(i=1; i < n-1; ++i) {
      if (dot(surf[i-1],surf[i]) > 0.5f) {
         pts[k] = raw[i];
         norm[k] = surf[i-1] + surf[i];
         norm[k].normalize();
         ++k;
      } else {
         pts[k] = raw[i];
         norm[k] = surf[i-1];
         ++k;
         pts[k] = raw[i];
         norm[k] = surf[i];
         ++k;
      }
   }
   pts[k] = raw[i];
   norm[k] = vec(1,0,0);
   ++k;   
   n = k;

   m = new Mesh;
   for (i=0; i < n; ++i) {
       vec p = pts[i];
       p.z = 0;
       initial[i] = m->addVertex(p, norm[i],pts[i].z,0);
   }

   memcpy(prev, initial, sizeof(prev));

   for (j=1; j <= slices; ++j) {
       float t_cur = (float) j / slices;
       float alpha = t_cur * 3.141592 * 2;
       float s = sin(alpha), c = cos(alpha);
       if (j == slices) s = 0, c = 1;
       for (i=1; i < n-1; ++i) {
           // rotate y value
           vec p,n;
           p.x = pts[i].x;
           p.y = c * pts[i].y;
           p.z = s * pts[i].y;
           n.x = norm[i].x;
           n.y = c * norm[i].y - s * norm[i].z;
           n.z = s * norm[i].y + c * norm[i].z;
           cur[i] = m->addVertex(p,n,pts[i].z,t_cur);
       }
       m->addTriangle(initial[0], cur[1], prev[1]);
       for (i=1; i < n-2; ++i) {
           m->addTriangle(cur[i], cur[i+1], prev[i]);
           m->addTriangle(cur[i+1],prev[i+1],prev[i]);
       }
       m->addTriangle(cur[i], initial[n-1], prev[i]);
       memcpy(prev, cur, sizeof(cur));
   }
   return m;
}


