// mesh/model stuff

struct Vertex
{
   vec loc;
   vec norm;
   float s,t;
};

struct Mesh
{
   int display_list;
   
   vector<Vertex> vertices;
   vector<int> indices;

   int addVertex(const vec &loc, const vec &normal, float s, float t);
   // {
   //   Vertex v; v.loc=loc; v.s=s; v.t=t;
   //   int n = vertices.size();
   //   vertices.push_back(v);
   //   indices.push_back(n);
   //}

   void addCylinder(const vec &p0, float r0, const vec &p1, float r1, int n);

   void addTriangle(int a, int b, int c) {
      indices.push_back(a);
      indices.push_back(b);
      indices.push_back(c);
   }

   void draw();
   
   void optimize();
};

extern Mesh *revolute_surface(char **src, int n_src, float scale_x, float scale_y, int slices);
// defined in artwork.cpp
extern Mesh *get_mesh(char *name);
extern GLuint get_texture(char *name);
extern Model *get_model(char *name);


struct Transform
{
   mat ori;
   vec loc;
};

struct Model
{
   vector<Mesh *> meshes;
   vector<GLuint> textures;
   vector<Transform> xforms;

   void addMesh(Mesh *m, GLuint tex, Transform *t) {
      if (m == NULL) return;
      meshes.push_back(m);
      textures.push_back(tex);
      xforms.push_back(*t);
   }

   void addMesh(Mesh *m, GLuint tex) {
      if (m == NULL) return;
      Transform t;
      t.loc = vec();
      t.ori.identity();
      addMesh(m, tex, &t);
   }
   
   void addMesh(char *mesh, char *tex) {
      addMesh(get_mesh(mesh), get_texture(tex));
   }

   void addMesh(char *mesh, char *tex, Transform *t) {
      addMesh(get_mesh(mesh), get_texture(tex), t);
   }

   void draw();
};

