#include "s.h"

vec r3vec(float z) { return rvec(z,z,z); }

void explosion(const vec &where, int num, float speed, float life, float z_bias, int subtype)
{
if (num > 25) num=25;
   for (int i=0; i < num; ++i) {
      Object *o = new Object(where, T_effect);
      o->subtype = subtype;
      o->velocity = r3vec(speed);
      o->velocity.x /= LANE_SPACING;
      o->velocity.z += z_bias;
      o->lifetime = life;
      worldAdd(o);
   }
}

void sparks(const vec &where, int num, vec speed, float life, float range, int subtype)
{
   for (int i=0; i < num; ++i) {
      Object *o = new Object(where, T_effect);
      o->subtype = subtype;
      o->velocity = speed;
      o->velocity += range * vec(frand(),frand(),frand());
      o->velocity.x /= LANE_SPACING;
      o->lifetime = life;
      worldAdd(o);
   }
}

extern void death_sound(Object *o);
void Object::die()
{
   // can't die more than once
   if (destroyed) return;

   explosion(loc, 30, 15, .75, -10, T_effect);
   death_sound(this);
   if (type == T_player) {
      health = 0;
      speed  = 0;
   } else {
      kill();
   }
}

mat Object::orientation()
{
   mat m;
   m =   rotationMatrix(angles.z, vec(0,0,1))
       * rotationMatrix(angles.x, vec(1,0,0))
       * rotationMatrix(angles.y, vec(0,1,0));
   return m;
}

void drawCone(float r, float ht, int count, vec *normals)
{
   static float s[5], c[5];
   float d = ht > 0 ? 1 : -1;
   if (c[0] == 0) {
      for (int i=0; i < 4; ++i) {
         float a = i * 6.283184 / 4;
         s[i] = sin(a);
         c[i] = -cos(a);
      }
      s[i] = s[0];
      c[i] = c[0];
   }
   glBegin(GL_TRIANGLE_FAN);
      glVertex3f(0,0,ht);
      if (normals) glNormal3fv(&normals[0].x);
      if (count == 4) {
         if (d < 0)
            for (int i=4; i >= 0; --i) {
               if (normals) glNormal3fv(&normals[i].x);
               glVertex3f(s[i]*r, c[i]*r, 0);
            }
         else
            for (int i=0; i <= 4; ++i) {
               if (normals) glNormal3fv(&normals[i].x);
               glVertex3f(s[i]*r, c[i]*r, 0);
            }
      } else {
         for (int i=0; i <= count; ++i) {
            float a = d*i * 6.283184 / count;
            glVertex3f(sin(a)*r, -cos(a)*r, 0);
         }
      }
   glEnd();
}

static float c_a;
vec Object::color()
{
   c_a = 0.5;
   switch (type) {
      case T_base  : return vec(.25,.25,.25);
      case T_player: return vec(0,0,1);
      case T_object: return vec(1,0,0);
      case T_bullet:
         if (health == 50) return vec(1,1,0.5);
         if (health == 10) return vec(0.25,0.5,1);
         if (health == 11) { c_a = 0.2; return vec(0.25,0.25,0.8); }
         return vec(0.55,0.4,0.65);
      case T_effect: 
         if (subtype == E_particle) return vec(1,1,0.5);
         if (subtype == E_bomb) return vec(0.4,1,0.3);
         if (subtype == E_soul) return vec(1.0,0.2,0.2);
         if (subtype == E_soul2) return vec(0.8,0.2,0.4);
         return vec(.5,.5,.5);
      default: assert(0); return vec();
   }
}

vec cone_normals[10] = {
   vec(-1,-1, 1),
   vec( 1,-1, 1),
   vec( 1, 1, 1),
   vec(-1, 1, 1),
   vec(-1,-1, 1),

   vec( 1,-1,-1),
   vec( 1, 1,-1),
   vec(-1, 1,-1),
   vec(-1,-1,-1),
   vec( 1,-1,-1),
};

extern float powerup_left[], powerup_start[];
extern Object *ghost;
void Object::draw()
{
   float temp[16];
   mat x;
   x.identity();
   vec l2;
   l2 = loc;
   l2.x = l2.x * LANE_SPACING + BAND_WIDTH/2 + LANE_SPACING/2;
   if (l2.z == 0) l2.z = compute_map_height(l2.x, l2.y);

   glColor3f(1,1,1);

   if (type == T_player) {
      float s = cos(music_time*3);
      float c = sin(music_time*6);
      l2.x += 0.2 * s;
      l2.z += 0.1 * c;

      if (this == ghost) {

         if (powerup_left[1] < -1) return;

         float t = (cos(music_time * 4)+1)/2;
         float a = 0.1 + 0.2 * t;
         float b = 0.3 + 0.7 * t;
         if (powerup_left[1] < 0) {
            a *= powerup_left[1] + 1;
            b *= powerup_left[1] + 1;
         }
         if (powerup_start[1]*2 < 1) {
            a *= powerup_start[1]*2;
            b *= powerup_start[1]*2;
         }
         glColor4f(b,b,b, a);
      }
   }

   glmat(temp, orientation(), l2);
   glMatrixMode(GL_MODELVIEW);
   glPushMatrix();
   glMultMatrixf(temp);
   const float zero[4] = { 0,0,0,0 };
   glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, zero);

   if (model) {
      glEnable(GL_TEXTURE_2D);
      glEnable(GL_LIGHTING);
      if (this == ghost) {
         glEnable(GL_BLEND);
         glDisable(GL_DEPTH_TEST);
         glDepthMask(GL_FALSE);
      }
      if (type == T_object && subtype == E_powerup) {
         const float powerc[4] = { 0.25,0.25,1 };
         if (subsubtype == 1)
            glColor3f(0.75,0.25,0.75);
         else
            glColor3f(0.25,0.25,0.5);
         glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, powerc);
      }
      model->draw();
      if (this == ghost) model->draw(); // draw twice for overbrightening
      glDisable(GL_BLEND);
      glEnable(GL_DEPTH_TEST);
      glDisable(GL_LIGHTING);
      glDepthMask(GL_TRUE);
   } else {
      glDisable(GL_TEXTURE_2D);
      vec v = color();
      glColor4f(v.x,v.y,v.z, c_a);
      glEnable(GL_BLEND);
      glDepthMask(GL_FALSE);
      glDisable(GL_DEPTH_TEST);
      if (type != T_effect) glEnable(GL_LIGHTING);
      if (cone_normals[0].x == 1)
         for (int i=0; i < 10; ++i)
             cone_normals[i].normalize();
      glShadeModel(GL_FLAT);
      drawCone(radius(),  radius(), 4, cone_normals+0);
      drawCone(radius(), -radius(), 4, cone_normals+5);
      glShadeModel(GL_SMOOTH);
      glDisable(GL_LIGHTING);
      glDisable(GL_BLEND);
      glDepthMask(GL_TRUE);
      glEnable(GL_DEPTH_TEST);
   }
   glPopMatrix();
   #ifdef DISPLAY_VELOCITY_LINES
   glColor3f(1,1,1);
   glBegin(GL_LINES);
   glVertex3fv(&loc.x);
   vec end = loc + velocity;
   glVertex3fv(&end.x);
   glEnd();
   #endif
}

#define MAX_DELTAV_PER_SEC    500
void Object::tick(float dt)
{
   reload_time -= dt;

   vec old_loc = loc;

   angles += turn_rate * dt;

   moveTo(loc + dt*velocity);

   if (destroyed) return;

   if (type == T_object)
      aiTick(this, dt);

   if (type == T_bullet || type == T_effect) {
      lifetime -= dt;
      if (lifetime < 0 || (type == T_bullet && loc.y >= LANE_LENGTH)) {
         if (type == T_effect && subtype == E_bomb) {
            explosion(loc, 40, 90, 1.1, 60, E_soul);
            score -= 4000000;
         }   
         kill();
      }
   }
}

bool sphereIntersect(const vec &a, const vec &b, float radius_sum)
{
   // determine if object a and object b are touching
   // assume they're spheres
   vec d = a - b;
   return dot(d,d) < radius_sum * radius_sum;
}

float linesegSphereIntersect(const vec &start, const vec &end, const vec &sphere_center, float rad)
{
   vec base  = start - sphere_center;
   vec delta = end   - start;
   
   // p(t) = base + t*delta
   // find t for which ||p(t)|| = rad

   // define [x] = dot(x,x)
   // so, find t for which [ p(t) ] = rad*rad

   // [ p(t) ] = [ base + t*delta ]
   //          = [ base ] + 2t*(base . delta) + t*t*[delta]
   //       r2 = [ base ] + 2t*(base . delta) + t*t*[delta]
   
   // now use the altered quadratic formula:
   //   At^2 + 2Bt + C = 0
   //   => t = ( -B +- sqrt(B^2 - AC) )/A

   float A = dot(delta,delta);
   if (A == 0) return -1;

   float B = dot(base,delta);
   float C = dot(base,base) - rad*rad;

   float det = B*B - A*C;
   if (det <= 0) return -1;

   // we want the smaller t that it hits (even if that's negative)
   // A is positive, so the +- should be -

   float t = (-B - sqrt(det))/A;
   if (t > 1) return -1;

   return t;
}

bool sweepIntersect(Object *a, vec goal, Object *b, vec *limit)
{
    assert (a != b);

    float radius_sum = a->radius() + b->radius();
   
    if (!(fabs(a->loc.x - b->loc.x) < 0.1f && fabs(goal.x - b->loc.x) < 0.25f)) return false;
    if (fabs(goal.z - b->loc.z) > radius_sum) return false;
    float m,n;
    m = a->loc.y;
    n = goal.y;
    if (m > n) { float t=m; m=n; n=t; }
    // if 
    m -= radius_sum;
    n += radius_sum;
    // check if m..n brackets b->y
    if (!(m <= b->loc.y && n >= b->loc.y))
        return false;

   //float t = linesegSphereIntersect(a->loc, goal, b->loc, radius_sum);
   //if (t < 0 || t > 1) return false;

   float t = 0.5f;
   //t -= 0.002f;
   //if (t < 0) t = 0;
   *limit = a->loc + (goal - a->loc) * t;
   return true;
}

static bool check_collisions(Object *a)
{
   if (a->type == T_object && a->subtype == E_powerup) return false;
   if (a->type == T_player) return false;
   return a->type != T_effect;
}

static bool considerCollisionPair(Object *a, Object *b)
{
   if (a == b) return false;
   if (a->creator == b) return false;
   if (b->creator == a) return false;
   if (a->type == T_bullet && b->type == T_bullet) {
      if (a->velocity.y * b->velocity.y > 0)
         return false;
   }
   return true;
}

void Object::moveTo(const vec &goal)
{
   // move from 'loc' to 'goal'...
   vec dest = goal;
   if (check_collisions(this)) {
      for (int i=0; i < objs.size(); ++i) {
         if (!objs[i]->destroyed && check_collisions(objs[i])) {
            if (considerCollisionPair(this, objs[i])) {
               vec limit;
               if (sweepIntersect(this, dest, objs[i], &limit)) {
                  if (collide(this, objs[i], (objs[i]->loc - limit).normal())) {
                     // if return true, then don't allow interpentration
                     // move object 0.01 unit from interpenetration
                     vec d = (objs[i]->loc-limit).normal();
                     dest = limit - d * 0.1f;
                  }
                  // if this got destroyed, bail
                  if (destroyed)
                     return;
               }
            }
         }
      }
   }
   loc = dest;
}

bool tt(ObjType s, Object *a, ObjType t, Object *b)
{
   return a->type == s && b->type == t;
}

bool collide(Object *a, Object *b, const vec &a_to_b_normal)
{
   vec n = a_to_b_normal;
   assert (a != b);

   // sort types to simplify computation
   if (a->type > b->type) {
      swap(a,b);
      n *= -1;
   }

   // types:
   //    T_player
   //    T_object
   //    T_bullet

   // bullet vs. bullet
   if (a->type == T_bullet) {
      a->kill();
      b->kill();
      return true; // interrupt movement
   }

   // bullet vs. non-bullet
   if (b->type == T_bullet) {
      if (a->health) {
         a->health -= b->health; 
         if (a->health < 1) {
            aiDie(a);
            a->die();
         } else {
            explosion(b->loc, 10, 10, 0.5, 0, T_effect);
         }
      }
      b->kill();
      return false;
   }

   return false;
}

float Object::radius()
{
   switch (type) {
      case T_base  :
      case T_player:
      case T_object: return  v->radius;
      case T_bullet:
         if (health < 10) return 0.7;
         if (health > 20) return 1.3;
         return  1;
      case T_effect:
         if (subtype == E_particle) return 0.15;
         if (subtype == E_bomb) return 0.75;
         if (subtype == E_soul) return 3.5;
         if (subtype == E_soul2) return 0.15;
         return 0.25;
      default: assert(0); return 0;
   }
}
