#include "s.h"
#include "stb_image.h"

#define GL_CLAMP_TO_EDGE 0x812F

#define parm2d(a,b) glTexParameteri(GL_TEXTURE_2D,a,b)
void texprep()
{
   parm2d(GL_TEXTURE_WRAP_S, GL_REPEAT);
   parm2d(GL_TEXTURE_WRAP_T, GL_REPEAT);
   parm2d(GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   parm2d(GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
}

void textureClamp(GLuint tex)
{
   glBindTexture(GL_TEXTURE_2D, tex);
   parm2d(GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   parm2d(GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
}

void textureClampS(GLuint tex)
{
   glBindTexture(GL_TEXTURE_2D, tex);
   parm2d(GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
   parm2d(GL_TEXTURE_WRAP_T, GL_REPEAT);
}

GLuint texCreate(int w, int h, uint8 *data)
{
   GLuint tex;
   glGenTextures(1, &tex);
   glBindTexture(GL_TEXTURE_2D, tex);
   gluBuild2DMipmaps(GL_TEXTURE_2D,GL_RGBA8,w,h,
                 GL_RGBA,GL_UNSIGNED_BYTE,data);
   texprep();
   memset(data,0,sizeof(data));
   return tex;
}

GLuint textureLoad(char *file)
{
   GLuint tex;
   int w,h,n;
   uint8 *data = stbi_load(file, &w, &h, &n, 4);
   if (!data) {
      char filename[256];
      sprintf(filename, "data/%s", file);
      data = stbi_load(filename, &w, &h, &n, 4);
      if (!data)
         return 0;   
   }
   
   tex = texCreate(w,h,data);
   free(data);
   return tex;   
}

GLuint textureAlpha(char *file)
{
   GLuint tex;
   int w,h,n,i;
   uint8 *data = stbi_load(file, &w, &h, &n, 4);
   if (!data) {
      char filename[256];
      sprintf(filename, "data/%s", file);
      data = stbi_load(filename, &w, &h, &n, 4);
      if (!data)
         return 0;   
   }
   if (n == 3 || n == 1) {
      for (i=0; i < 4*w*h; i += 4)
         data[i+3] = data[i+0];
   }
   for (i=0; i < 4*w*h; i += 4)
      data[i+0] = data[i+1] = data[i+2] = 255;
   
   tex = texCreate(w,h,data);
   free(data);
   return tex;   
}

// texture synthesizer

// "plasma fractal" generator instead of perlin noise
// just because it's simpler to code and control

float fdata[FSIZE][FSIZE];
static uint8 data[FSIZE][FSIZE][4];

static void plasma(int s, float delta) {
   int s2 = s >> 1;
   for (int y0 = 0; y0 < FSIZE; y0 += s)
   for (int x0 = 0; x0 < FSIZE; x0 += s) {
      int x1 = (x0+s) & (FSIZE-1), y1 = (y0+s) & (FSIZE-1);
      float a = fdata[y0][x0];
      float b = fdata[y0][x1];
      float c = fdata[y1][x0];
      float d = fdata[y1][x1];
      fdata[y0   ][x0+s2] = (a+b)/2 + frand()*delta;
      fdata[y0+s2][x0   ] = (a+c)/2 + frand()*delta;
      fdata[y0+s2][x0+s2] = (a+b+c+d)/4 + frand()*delta;
   }
}

void texFractal(int start, float delta, float decay) {
   for(int s = FSIZE >> start; s > 1; s >>= 1) {
      plasma(s, delta);
      delta *= decay;
   }
}

static void add(int x, int y, int p, int v) {
   int z = data[y][x][p];
   data[y][x][p] = clamp(z+v,0,255);
}

void texScale(int r1, int g1, int b1, int r2, int g2, int b2)
{
   for (int j=0; j < FSIZE; ++j)
   for (int i=0; i < FSIZE; ++i) {
      float p = clamp(fdata[j][i],-1,1);
      p = (p+1)/2;
      add(i,j,0, r1 + p*(r2-r1));
      add(i,j,1, g1 + p*(g2-g1));
      add(i,j,2, b1 + p*(b2-b1));
   }
   memset(fdata, 0, sizeof(fdata));
}

void texCheck(int x, int y, float r, float g, float b)
{
   for (int j=0; j < FSIZE; ++j)
   for (int i=0; i < FSIZE; ++i)
      if (((j / y) + (i / x)) & 1) {
         data[j][i][0] *= r;
         data[j][i][1] *= g;
         data[j][i][2] *= b;
      }
}

GLuint texMake()
{
   GLuint tex = texCreate(FSIZE,FSIZE,data[0][0]);
   memset(data,0,sizeof(data));
   return tex;
}

