//   N bins
//   M bins
//      1 = 2048  2^11
//      2^(1/k) == 11k + 1 bins
//   M bins: k = M/11
//      2^(11/M)
//      1/log2(M/11)


#include "s.h"
#include "midi.h"

bool paused = false;
bool menu = true;

bool dont_sim(void)
{
   return (paused || menu);
}

GLuint earth;

int NUM_LANES=10;
float LANE_SPACING=5;
int player_min=3;
int player_max=10;
int num_player=7;

void set_num_lanes(int n, int p)
{
    NUM_LANES = n;
    LANE_SPACING = 50.0f / n;
    player_min = p;
    player_max = n;
    num_player = n-p;
}

Vehicle player_ship = {   1, 100,300,    1,   2,    2,  5, .5, 200 };
Vehicle turret      = {  10,   0,  0,    9,   9,   10, 10,  2, INFINITE_MASS };
Vehicle base        = { 2.5,   0,  0,    0,   0,    0,  0,  0, INFINITE_MASS };

vector<Object *> objs;
Object *player, *p2, *mbase, *ghost;

vector<Base *> bases;

void worldAdd(Object *o) {
   objs.push_back(o);
}

GLuint floor_tex;
int scr_w, scr_h;

//Mesh *test;
//Model *test_model;

void gameInit() {
   menu = false;
   aiInit();
}

void worldInit() {
   set_num_lanes(12, 4);
   floor_tex = textureLoad("lines.png");
   textureClampS(floor_tex);

   mapInit();
   textInit();

   init_artwork();

   player = new Object(vec(-100,0,0), T_player);
   player->v       = &player_ship;
   player->model = get_model("tube");
   
   worldAdd(player);
   p2 = new Object(vec(-100,0,0), T_player);
   p2->v = &player_ship;
   p2->model = get_model("cannon");
   worldAdd(p2);

   ghost = new Object(vec(5,-8,0), T_player);
   ghost->v = &player_ship;
   ghost->model = get_model("ghost");
   worldAdd(ghost);
   
   mbase = new Object(vec(1,0,0), T_player);
   mbase->v = &base;
   mbase->model = get_model("base");
   worldAdd(mbase);
   
   earth = textureLoad("earth2.jpg");

   //test = revolute_surface(surf, sizeof(surf)/sizeof(surf[0]), 0.125, 6);
   //test->optimize();
   //test_model = new Model;
   //test_model->addMesh(test, floor_tex);
   
   
}

vec camera_location;
Object *camera_object;
void setCamera(Object *cam)
{
   mat m = cam->orientation().transpose();
   glRotatef(-90, 1,0,0);

   float temp[16];
   glmat(temp, m, m * cam->loc * -1);
   glMultMatrixf(temp);

   camera_object = cam;
   camera_location = cam->loc;
}

void fireShot(Object *obj, float speed, float damage, int lane)
{
   assert(lane >= 0);
   
   // create a bullet at the object's location
   Object *bullet = new Object(obj->loc, T_bullet);
   bullet->loc.x = lane;
   bullet->creator = obj;
   bullet->health = damage;

   // bullet goes in direction object faces
   vec dir = obj->orientation() * vec(0,1,0);
   bullet->velocity = dir * speed;
   bullet->velocity += obj->velocity;

   bullet->loc += dir * (obj->radius()/2 + bullet->radius()*1.15f);
   bullet->turn_rate.y = 4 + (damage/3) + 3 * prand();
   if (damage < 10) bullet->turn_rate *= -1;
   if (damage == 1) bullet->turn_rate.z = 7;
   //bullet->loc += obj->orientation() * vec(3,0,0);

   // how long before bullet self-destructs?
   bullet->lifetime = 4.0f;
   obj->reload_time = obj->v->reload_delay;
   worldAdd(bullet);
}


static int sw,sh,mx,my;

#define key(x)  glfwGetKey(x)
#define skey(x) glfwGetKey(GLFW_KEY_##x)

float powerup_left[4] = { -9,-9,-9,-9 }, powerup_start[4] = { 9,9,9,9 };

extern int player_lane;
extern float player_lane_frac;
void playerTick(float dt)
{
   for (int i=0; i < 4; ++i) {
      powerup_start[i] += dt;
      powerup_left[i] -= dt;
   }
    player_lane = player_min + (int) ((mx * num_player) / sw);
    if (player_lane < player_min) player_lane = player_min;
    if (player_lane >= player_max) player_lane = player_max-1;
    player->loc.x = (float) num_player * mx / sw + player_min - 0.5;
    player->loc.y = -4;
    player->loc.z = 0;
    player_lane_frac = player->loc.x + 0.5 - player_lane;
    if (player_lane_frac < 0) player_lane_frac = 0;
    if (player_lane_frac > 1) player_lane_frac = 1;
    p2->loc.x =  player_lane;
    p2->loc.y = -2.25;
    p2->loc.z = 0;
}

static void texRect(float s, float t, vec a, vec b, vec c, vec d)
{
   glEnable(GL_TEXTURE_2D);
   glBegin(GL_POLYGON);
      glTexCoord2f(0,0); glVertex3fv(&a.x);
      glTexCoord2f(0,s); glVertex3fv(&b.x);
      glTexCoord2f(t,s); glVertex3fv(&c.x);
      glTexCoord2f(t,0); glVertex3fv(&d.x);
   glEnd();
   glDisable(GL_TEXTURE_2D);
}

void rect2dtex(float s0, float t0, float s1, float t1, float x, float y, float dx, float dy)
{
   glBegin(GL_POLYGON);
      glTexCoord2f(s0,t0); glVertex2f(x,y);
      glTexCoord2f(s0,t1); glVertex2f(x,y+dy);
      glTexCoord2f(s1,t1); glVertex2f(x+dx,y+dy);
      glTexCoord2f(s1,t0); glVertex2f(x+dx,y);
   glEnd();
}

void rect2dtex_stripped(float s0, float t0, float s1, float t1, float x, float y, float dx, float dy)
{
   glTexCoord2f(s0,t0); glVertex2f(x,y);
   glTexCoord2f(s0,t1); glVertex2f(x,y+dy);
   glTexCoord2f(s1,t1); glVertex2f(x+dx,y+dy);
   glTexCoord2f(s1,t0); glVertex2f(x+dx,y);
}

void rect2dtex_fat(float s0, float t0, float s1, float t1, float x, float y, float dx, float dy)
{
   int i,j;
   glBegin(GL_QUADS);
   for (j = -2; j <= 2; j += 2) {
      for (i = -2; i <= 2; i += 2) {
         if (i || j)
            rect2dtex_stripped(s0,t0,s1,t1, x+i,y+j, dx,dy);
      }
   }
   glEnd();
}

void rect2d(float x, float y, float dx, float dy) {
   rect2dtex(0,0,0,0,x,y,dx,dy);
}

float last_frame_time;
double score = 10000000000.0;

char *comma_score(double score)
{
   double sc = score;
   static char output[64];
   int doit=0;
   char *s = output;
   char *fmt = "%3d,";
   if (sc > 999999999) {
      int z = floor(sc/1000000000.);
      sprintf(s, fmt, z);
      sc -= 1000000000.*z;
      fmt = "%03d,";
      doit = 1;
   } else
      strcpy(s, "   ");
   s += strlen(s);
   if (doit || sc > 999999) {
      int z = floor(sc/1000000.);
      sprintf(s, fmt, z);
      sc -= 1000000.*z;
      fmt = "%03d,";
      doit = 1;
   } else
      strcpy(s, "   ");
   s += strlen(s);
   if (doit || sc > 999) {
      int z = floor(sc/1000.);
      sprintf(s, fmt, z);
      sc -= 1000.*z;
      fmt = "%03d,";
      doit = 1;
   } else
      strcpy(s, "   ");
   s += strlen(s);
   int z = sc;
   sprintf(s, fmt, z);
   *strchr(s,',') = 0;
   return output;
}

#define S(t,s)       { t,s,NULL,0 }
#define S2(t,s1,s2)  { t,s1,s2,0  }
#define SP(t,s)       { t,s,NULL,1 }
#define SP2(t,s1,s2)  { t,s1,s2,1  }

struct
{
   float t;
   char *str1;
   char *str2;
   int split;
} tuts[] =
{
   S(  0, "Move mouse sideways"),
   S2(  4, "","ADB             "),
   S(  5.5, "Automatic Defense Base"),
   S(  9, "Base defends green lanes"),
   S2( 11, "", "You defend blue lanes"),
   S( 16, "Mouse buttons fire"),
   SP( 20, "Left = power shot"),
   SP( 24, "Kills most enemies"),
   SP2( 30, "Hold down for", "auto-repeat fire"),

   S2( 38, "Power shot uses", "humans as ammo"),
   S2( 43, "Right mouse button", "fires weak shot"),
   S2( 48, "Ineffective against", "most targets"),
   S2( 53, "Use against", "tiny invaders"),
   SP2( 58, "Weak shot does NOT", "consume humans"),
   SP2( 66, "Hold down right", "for rapid weak shot"),

   S2( 79, "Enemies have", "varying abilities"),
   SP2(83, "Spinners drop more bombs", "as they get closer"),
   SP2(92, "Minimize casualties", "by shooting in order"),
   SP2(99, "Collect powerups by", "colliding with them"),
   S(112, "GOOD LUCK"),
   S(200,NULL),
};
#undef S
#undef S2

void display(float scale, float x, float y, char * a, float p)
{
   glColor4f(1,1,1,p);
   print2(scale,x,y, a);
   glColor4f(.2,.2,1,p);
   print(scale,x,y, a);
}

extern bool show_tutorial;
void tutorial(double t)
{
   int i;
   if (!show_tutorial) return;

   if (t > tuts[ACOUNT(tuts)-1].t) return;
      
   for (i=0; i < ACOUNT(tuts); ++i) {
      if (t < tuts[i].t) continue;
      if (t > tuts[i].t + 12) continue;
      char *a = tuts[i].str1;
      char *b = tuts[i].str2;
      if (tuts[i].str1 == NULL) continue;
      float dt = t - tuts[i].t;
      float show_time = 2;
      if (b) show_time += 1.5;
      
      float fade_time = 1;
      if (dt > show_time + fade_time) continue;
      
      float p = 1;
      if (dt > show_time) p = 1 - (dt - show_time) / fade_time;
      if (p <= 0) continue;
      
      int w1 = strlen(a);
      int w2 = b ? strlen(b) : 0;
      
      w1 *= 16*2.5;
      w2 *= 16*2.5;

      //p = 1 - (1*p)*(1-p);
      p *= 0.75;
      p *= 1-0.5*(sin(music_time * 4)+1)/2;
      assert(p >= 0 && p <= 1);
      
      float s = 1 + (sin(music_time * 3)/20);
      float q = music_time / 8;
      float dx = 6 * sin(q);
      float dy = 4 * cos(q);
      
      w1 *= s;
      w2 *= s;
      
      int v1 = scr_h-200;
      int v2 = scr_h - 300;
      
      if (tuts[i].split) { v1 += 50; p *= 0.8; }

      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);   
      display(2.5*s, dx+scr_w/2 - w1/2, dy+v1, a, p);
      if (b) {
         if (tuts[i].split) p *= 0.7;
         display(2.5*s,dx+scr_w/2 - w2/2, dy+v2, b, p);
      }
   }
}

#define NUM_MENU_CHOICES  3
int menu_choice;
int menu_pos_x, menu_pos_y;

void show_menu(void)
{
   char *choices[NUM_MENU_CHOICES] = { 
      "Tutorial",
      "Play",
      "Exit",
   };
   float spacing[NUM_MENU_CHOICES] = { 7.5,4,4 };
   
   int i;
   for (i=0; i < NUM_MENU_CHOICES; ++i) {
      float size = (i == menu_choice) ? 4.5 : 3.8;
      display(size, scr_w/2 - spacing[i]*size*8 - size*8, scr_h*3/4 - scr_h/2*i/NUM_MENU_CHOICES + (size-3)*16, choices[i], 0.8);
   }
   
   glColor4f(.8,1,.8,0.7);
   print(1, menu_pos_x, menu_pos_y, "x");
}

extern char debug_string[];
void worldDraw() {
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();

   glfwGetWindowSize(&scr_w,&scr_h);
   gluOrtho2D(0,scr_w,0,scr_h);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   glDisable(GL_CULL_FACE);
   glDisable(GL_DEPTH_TEST);
   glDisable(GL_FOG);

   glColor3f(1,1,1);
   glEnable(GL_TEXTURE_2D);
   if (earth) {
       if (menu)
          glColor3f(0.6,0.6,0.6);
       else
          glColor3f(0.3f,0.3f,0.3f);
       glBindTexture(GL_TEXTURE_2D, earth);
       rect2dtex(0,1,1,0,0,-220,1024,512);
   }
   
   glEnable(GL_DEPTH_TEST);
   
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(74, 1.33f, 0.25, 20000);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glRotatef(25,1,0,0);
   glTranslatef(-(NUM_LANES*LANE_SPACING+BAND_WIDTH)/2.0,-20,-22);
   glRotatef(-90, 1,0,0);

   glEnable(GL_CULL_FACE);

   glFogi(GL_FOG_MODE, GL_EXP);
   glFogf(GL_FOG_DENSITY, .0001f);
   float fog_color[4] = { 0,0,1,0 };//{ .3f, .3f, .6f };
   glFogfv(GL_FOG_COLOR, fog_color);
   glEnable(GL_FOG);

    GLfloat const LightPosition[4] = {1.0f, -0.5f, 0.5f, 0.0f};
    GLfloat const DiffuseColor[4] = {1.0f, 1.0f, 1.0f, 1.0f};
    GLfloat const SpecularColor[4] = { .7,.7,.7,1 };
    glLightfv(GL_LIGHT0, GL_POSITION, LightPosition);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, DiffuseColor);
    glEnable(GL_LIGHT0);
    glEnable(GL_COLOR_MATERIAL);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 8);

   if (!menu) mapDraw();
   glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
   glDisable(GL_BLEND);
   if (!menu) {
      static int first=1;
      if (!first)
         for(int i=0; i < objs.size(); ++i)
            objs[i]->draw();
      first = 0;
   }

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();

   glfwGetWindowSize(&scr_w,&scr_h);
   gluOrtho2D(0,scr_w,0,scr_h);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   glDisable(GL_CULL_FACE);
   glDisable(GL_DEPTH_TEST);
   glDisable(GL_FOG);

   glColor3f(1,1,1);

   glEnable(GL_BLEND);
   glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glColor4f(1,1,1,.55f);
   print(0.5, 10,20, " Frame rate %5.1fms/frame ", last_frame_time * 1000);
   print(0.5, 10,30, " Measure: %d  Time: %6.0lf", (int) (music_time / 0.2f / 8), music_time);
   print(0.5, 10,40, debug_string);
   
   glColor4f(1,1,1,.65f);
   if (!menu) {
      print(1.2f, scr_w/2 - 160, scr_h-10, "Earth Population");
      print(2,  scr_w/2 - 260, scr_h-40, comma_score(score));
   }
   
   if (menu)
      show_menu();
   else
      tutorial(music_time);
   if (powerup_start[0] < 2) {
      float s = 3.5+powerup_start[0];
      display(s, scr_w/2 - 16*s*9/2, scr_h/4, "TIME SLOW", (1-powerup_start[0]/2)/2);
   }
   if (powerup_left[0] < 2 && powerup_left[0] > 0) {
      float s = 3.5+powerup_left[0];
      display(s, scr_w/2 - 16*s*12/2, scr_h/4, "TIME RESTORE", (1-powerup_left[0]/2)/4);
      if (powerup_left[0] < 1 && powerup_left[0] + last_frame_time >= 1)
         play_note(11,64,127,1);
   }
   if (powerup_left[1] > 0 && powerup_start[1] < 2) {
      float s = 3.5+powerup_start[1];
      display(s, scr_w/2 - 16*s*14/2, scr_h/4, "SOUL HARVESTER", (1-powerup_start[1]/2)/2);
   }
}

static void objectsTick(float dt) {
   for(int i=0; i < objs.size(); ++i)
      objs[i]->tick(dt);
}

static void cleanupDeleted()
{
   for (int i=0; i < objs.size();)
      if (objs[i]->destroyed) {
         objs[i] = objs.back();
         objs.pop_back();
      } else
         ++i;
}

extern void updateMusic(float dt);

extern void update_ripples(float dt);
extern bool fullscreen;

void worldTick(float dt) {
   if (glfwGetKey('S'))
      dt = 0.001;
   if (glfwGetKey('F'))
      dt = 1;

   last_frame_time = dt;

   dt = clamp(dt,0,.125);

   glfwGetWindowSize(&sw,&sh);
   glfwGetMousePos(&mx,&my);
   
   if (menu) {
      if (mx < 10) mx = 10;
      if (mx > sw/2) mx = sw/2;
      if (fullscreen) {
         if (my < 0) my = 0;
         if (my > sh/5) my = sh/5;
         glfwSetMousePos(mx,my);
         // remap my so it's always within the menu
         my = my * 2 + sh/4;
         mx = mx * 2 - 10;
      }
      menu_pos_x = mx;
      menu_pos_y = sh - my;
      menu_choice = floor((float)(my - sh/4) * NUM_MENU_CHOICES / (sh/2.4));
      
      static bool state = false;
      bool last_state = state;
      state = glfwGetMouseButton(GLFW_MOUSE_BUTTON_LEFT) != 0;
      if (!state && last_state) {
         switch(menu_choice) {
            case 0: show_tutorial = true; gameInit(); break;
            case 1: show_tutorial = false; gameInit(); break;
            case 2: exit(0);
         }
      }
      return;
   }

   if (!dont_sim()) {
      if (fullscreen) {
         if (mx < 10) mx = 10;
         if (mx > sw-10) mx = sw-10;
      } else {
         if (mx < 150) mx = 150;
         if (mx > sw - 150) mx = sw - 150;
         my = sh/2;
      }
      #ifndef _DEBUG
      glfwSetMousePos(mx,my);
      #endif

      // rescale mx to give bigger dead region in windowed mode,
      // so that we don't click out of window too much
      if (!fullscreen)
         mx = floor(mx * 1.25f - 0.125*sw);
      if (mx < -100) mx = -100; else if (mx >= sw+100) mx = sw+100;

      update_ripples(dt);
      if (powerup_left[0] > 0) {
         float slow_dt = dt * 0.8;
         float q = clamp(powerup_start[0]/2, 0, 1);
         float p = clamp(powerup_left[0]/2, 0, 1);
         dt = dt + p*q * (slow_dt - dt);
      }
      playerTick(dt);
      updateMusic(dt);
      objectsTick(dt);
   }

   for (int i=0; i < NUM_LANES; ++i) {
      for (int j=0; j < 4; ++j) {
          float p = j / 4.0;
          vec color = vec(0,0,0.25);
          p = 1-p;
          if (i < player_min)
             color = vec(0,0.25,0.125);
          lane_colors[i][j][0] = color.x * p;
          lane_colors[i][j][1] = color.y * p;
          lane_colors[i][j][2] = color.z * p;
          lane_colors[i][j][3] = 0;
      }
   }
   cleanupDeleted();
   //score -= dt * (prand() * 0.5 + 0.5) * 5000000;
}
